from collections import Counter
import copy
import itertools
import random
from dataclasses import dataclass
from dataclasses import field
from typing import Optional

from Class.exceptions import (
    GeneratorException,
    CantAssignItemException,
    SettingsException,
)
from Class.itemClass import KH2Item
from Class.newLocationClass import KH2Location
from Class.randomUtils import weighted_sample_without_replacement
from List.ObjectiveList import KH2Objective
from List.ItemList import Items
from List.NewLocationList import Locations
from List.configDict import (
    ObjectivePoolOption,
    locationCategory,
    itemRarity,
    itemType,
    locationDepth,
    locationType,
    LevelUpStatBonus,
)
from List.inventory import (
    ability,
    consumable,
    form,
    growth,
    keyblade,
    magic,
    misc,
    proof,
    report,
    storyunlock,
)
from List.inventory.item import InventoryItem
from List.location import (
    simulatedtwilighttown as stt,
    weaponslot,
    hundredacrewood as haw,
    hollowbastion as hb,
    starting,
)
from Module.RandomizerSettings import RandomizerSettings
from Module.depths import ItemDepths
from Module.modifier import SeedModifier
from Module.weighting import LocationWeights


@dataclass
class ItemAssignment:
    location: KH2Location
    item: Optional[KH2Item]
    item2: Optional[KH2Item] = None

    def items(self) -> list[KH2Item]:
        """Returns the items (if any) for this assignment."""
        result = []
        if self.item is not None:
            result.append(self.item)
        if self.item2 is not None:
            result.append(self.item2)
        return result

    def __eq__(self, o: object) -> bool:
        if isinstance(o, ItemAssignment):
            return self.location == o.location
        elif isinstance(o, KH2Location):
            return self.location == o
        else:
            return NotImplemented


@dataclass
class WeaponStats:
    location: KH2Location
    strength: int
    magic: int

    def __eq__(self, o: object) -> bool:
        if isinstance(o, WeaponStats):
            return self.location == o.location
        elif isinstance(o, KH2Location):
            return self.location == o
        else:
            return NotImplemented


@dataclass
class LevelStats:
    location: KH2Location
    experience: int
    strength: int
    magic: int
    defense: int
    ap: int

    def __eq__(self, o: object) -> bool:
        if isinstance(o, LevelStats):
            return self.location == o.location
        elif isinstance(o, KH2Location):
            return self.location == o
        else:
            return NotImplemented


@dataclass
class FormExp:
    location: KH2Location
    experience: int

    def __eq__(self, o: object) -> bool:
        if isinstance(o, FormExp):
            return self.location == o.location
        elif isinstance(o, KH2Location):
            return self.location == o
        else:
            return NotImplemented


@dataclass
class SynthRequirement:
    synth_item: KH2Item
    amount: int


@dataclass
class SynthesisRecipe:
    location: KH2Location
    unlock_rank: int
    requirements: list[SynthRequirement] = field(default_factory=list)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, SynthesisRecipe):
            return self.location == o.location
        elif isinstance(o, KH2Location):
            return self.location == o
        else:
            return NotImplemented


def _find_location(
    location_name: str, locations: list[KH2Location]
) -> Optional[KH2Location]:
    return next((loc for loc in locations if loc.name() == location_name), None)


def _find_locations(
    location_names: list[str], locations: list[KH2Location]
) -> list[KH2Location]:
    return [loc for loc in locations if loc.name() in location_names]


class Randomizer:
    def __init__(self, settings: RandomizerSettings, progress_bar_vis: bool = False):
        if settings is None:
            raise SettingsException(
                "Invalid settings passed to randomize. Change settings and try again"
            )
        random.seed(settings.full_rando_seed)
        self.progress_bar_vis = progress_bar_vis
        self.regular_locations = Locations(settings, secondary_graph=False)
        self.reverse_locations = Locations(settings, secondary_graph=True)
        self.master_locations = (
            self.regular_locations if settings.regular_rando else self.reverse_locations
        )
        self.location_weights = LocationWeights(
            settings, self.regular_locations, self.reverse_locations
        )
        self.report_depths = ItemDepths(settings.reportDepth, self.master_locations)
        self.proof_depths = ItemDepths(settings.proofDepth, self.master_locations)
        self.story_depths = ItemDepths(settings.storyDepth, self.master_locations)
        self.promise_charm_depths = ItemDepths(settings.promiseCharmDepth, self.master_locations)
        self.yeet_the_bear = settings.yeetTheBear
        self.num_valid_locations = None
        self.num_available_items = None
        self.starting_item_ids: list[int] = []
        self.assignments: list[ItemAssignment] = []
        self.donald_assignments: list[ItemAssignment] = []
        self.goofy_assignments: list[ItemAssignment] = []
        self.weapon_stats: list[WeaponStats] = []
        self.level_stats: list[LevelStats] = []
        self.form_level_exp: list[FormExp] = []
        self.synthesis_recipes: list[SynthesisRecipe] = []
        self.shop_items: list[KH2Item] = []
        self.objectives: list[KH2Objective] = []
        self.assign_sora_items(settings)
        if progress_bar_vis:
            return
        self.assign_party_items()
        self.assign_weapon_stats(settings)
        self.assign_level_stats(settings)
        self.assign_form_level_exp(settings)

    def assign_form_level_exp(self, settings: RandomizerSettings):
        """Assigns experience values to each form level."""
        experience_values = {
            locationCategory.SUMMONLEVEL: settings.summon_exp(),
            locationCategory.VALORLEVEL: settings.valor_exp(),
            locationCategory.WISDOMLEVEL: settings.wisdom_exp(),
            locationCategory.LIMITLEVEL: settings.limit_exp(),
            locationCategory.MASTERLEVEL: settings.master_exp(),
            locationCategory.FINALLEVEL: settings.final_exp(),
        }
        for category, experience in experience_values.items():
            locations = self.master_locations.locations_for_category(category)
            for index, location in enumerate(locations):
                self.form_level_exp.append(FormExp(location, experience[index]))

    def assign_level_stats(self, settings: RandomizerSettings):
        """Assigns stat increases to each level."""
        locations = self.master_locations.locations_for_category(locationCategory.LEVEL)
        locations.sort(key=lambda x: x.LocationId)

        # All the stats are accumulated as we go
        acc_strength = 2
        acc_magic = 6
        acc_defense = 2
        acc_ap = 0
        acc_exp = 0
        stats_added = 0

        def add_stat(choice: tuple[LevelUpStatBonus, int]):
            nonlocal acc_strength
            nonlocal acc_magic
            nonlocal acc_defense
            nonlocal acc_ap

            nonlocal stats_added
            stats_added+=1

            chosen_stat, stat_increase = choice
            if chosen_stat == LevelUpStatBonus.STRENGTH:
                acc_strength += stat_increase
            elif chosen_stat == LevelUpStatBonus.MAGIC:
                acc_magic += stat_increase
            elif chosen_stat == LevelUpStatBonus.DEFENSE:
                acc_defense += stat_increase
            elif chosen_stat == LevelUpStatBonus.AP:
                acc_ap += stat_increase
            else:
                raise GeneratorException("We had a problem assigning stats to levels")

        def add_no_stats(_):
            return

        adder_function = add_no_stats if settings.vanilla_level_one else add_stat

        level_stat_pool = [(s[0], s[1]) for s in settings.level_stat_pool]
        stat_weights = [s[2] for s in settings.level_stat_pool]
        experience = settings.sora_exp()
        excluded_levels = settings.excluded_levels()
        double_stat_levels = settings.double_stat_levels()
        for index, location in enumerate(locations):
            if index != 0:
                stat_choices = weighted_sample_without_replacement(
                    population=level_stat_pool, weights=stat_weights, k=2
                )
                adder_function(stat_choices[0])
                if location.LocationId in excluded_levels and location.LocationId in double_stat_levels:
                    adder_function(stat_choices[1])
            # print(f"{location} has total {stats_added} stat growths")
            acc_exp += experience[index + 1] - experience[index]
            self.level_stats.append(
                LevelStats(
                    location, acc_exp, acc_strength, acc_magic, acc_defense, acc_ap
                )
            )

    def assign_weapon_stats(self, settings: RandomizerSettings):
        """Assigns stats to each weapon."""
        key_min = settings.keyblade_min_stat
        key_max = settings.keyblade_max_stat
        sora_average = (key_min + key_max) // 2

        for key in weaponslot.keyblade_slots():
            if (
                key.LocationId == weaponslot.LocationId.Pureblood
                and not settings.pureblood
            ):
                continue
            self.weapon_stats.append(
                WeaponStats(
                    key,
                    strength=random.randint(key_min, key_max),
                    magic=random.randint(key_min, key_max),
                )
            )
        for struggle_weapon in weaponslot.struggle_weapon_slots():
            self.weapon_stats.append(
                WeaponStats(struggle_weapon, strength=sora_average, magic=sora_average)
            )
        for staff in weaponslot.donald_staff_slots():
            self.weapon_stats.append(
                WeaponStats(
                    staff, strength=random.randint(1, 13), magic=random.randint(1, 13)
                )
            )
        for shield in weaponslot.goofy_shield_slots():
            self.weapon_stats.append(
                WeaponStats(shield, strength=random.randint(1, 13), magic=0)
            )

    def assign_party_items(self):
        """Assigns items to locations for party members."""
        donald_locations = Locations.all_donald_locations()
        for donald_ability in Items.donald_ability_list():
            random_location = random.choice(donald_locations)
            if self.assign_item(
                random_location, donald_ability, self.donald_assignments
            ):
                donald_locations.remove(random_location)

        goofy_locations = Locations.all_goofy_locations()
        for goofy_ability in Items.goofy_ability_list():
            random_location = random.choice(goofy_locations)
            if self.assign_item(random_location, goofy_ability, self.goofy_assignments):
                goofy_locations.remove(random_location)

    def apply_starting_items(self, settings: RandomizerSettings):
        """Populates the starting items list, based on settings for starting inventory, growth, etc."""
        self.starting_item_ids.extend(settings.starting_inventory_ids)

        starting_growth_abilities = SeedModifier.starting_growth(
            settings.starting_growth_option
        )
        self.starting_item_ids.extend(
            growth_ability.id for growth_ability in starting_growth_abilities
        )

        starting_reports = random.sample(
            report.all_reports(), k=settings.starting_report_count
        )
        self.starting_item_ids.extend(rpt.id for rpt in starting_reports)

        starting_unlocks = SeedModifier.starting_unlocks(
            mode=settings.starting_visit_mode,
            random_range=settings.starting_visit_random_range,
            specific_unlocks=settings.starting_unlocks_per_world
        )
        self.starting_item_ids.extend(unlock.id for unlock in starting_unlocks)

        if settings.tt1_jailbreak:
            self.starting_item_ids.append(Items.getTT1Jailbreak().Id)
        if settings.objective_rando:
            self.starting_item_ids.append(Items.objectiveReportItem().Id)

    def initial_item_pool(self, settings: RandomizerSettings) -> list[KH2Item]:
        """
        Builds the initial item pool, based on settings.
        Also adds the configured number of reports and visit unlocks to the shop.
        """
        item_pool = Items.getItemList(
            story_unlocking_rarity=settings.story_unlock_rarity
        )

        # Add any additional items to the pool based on settings
        if settings.statSanity:
            item_pool.extend(Items.getStatItems())
        if settings.fifty_ap:
            item_pool.extend(itertools.repeat(KH2Item(consumable.ApBoost), 50))
        if settings.promiseCharm:
            item_pool.append(Items.getPromiseCharm())

        # Remove various items from the pool based on settings
        item_ids_to_remove: list[int] = []
        item_types_to_remove: list[itemType] = []
        if not settings.pureblood:
            item_ids_to_remove.append(keyblade.Pureblood.id)
        if not settings.antiform:
            item_ids_to_remove.append(form.AntiForm.id)
        if not settings.include_maps:
            item_types_to_remove.append(itemType.MAP)
        if not settings.include_recipes:
            item_types_to_remove.append(itemType.RECIPE)
        if not settings.include_accessories:
            item_types_to_remove.append(itemType.ACCESSORY)
        if not settings.include_armor:
            item_types_to_remove.append(itemType.ARMOR)
        if not settings.include_munny_pouches:
            item_types_to_remove.append(itemType.MUNNY_POUCH)

        # Any starting items get removed from the item pool
        item_ids_to_remove.extend(self.starting_item_ids)

        def include_item(item: KH2Item) -> bool:
            return item.ItemType not in item_types_to_remove
        item_pool = list(filter(include_item, item_pool))

        # Remove these one at a time since we can have duplicate IDs for starting inventory
        # (at the very least, visit unlocks)
        for item_id_to_remove in item_ids_to_remove:
            item_to_remove = next((item for item in item_pool if item.Id == item_id_to_remove), None)
            if item_to_remove is not None:
                item_pool.remove(item_to_remove)

        # Reports, visit unlocks, and keyblades in the shop (these do affect the item pool)
        if settings.shop_reports > 0:
            report_pool = [i for i in item_pool if i.ItemType == itemType.REPORT]
            num_reports_in_shop = min(settings.shop_reports, len(report_pool))
            chosen_reports: list[KH2Item] = random.sample(
                report_pool, k=num_reports_in_shop
            )
            self.shop_items.extend(chosen_reports)
            for chosen_report in chosen_reports:
                item_pool.remove(chosen_report)
        if settings.shop_unlocks > 0:
            # make the visit unlock pool unique so we only select a visit unlock once for the shop
            visit_unlock_pool = list(set([
                i for i in item_pool if i.ItemType == itemType.STORYUNLOCK
            ]))
            num_visit_unlocks_in_shop = min(
                settings.shop_unlocks, len(visit_unlock_pool)
            )
            chosen_unlocks: list[KH2Item] = random.sample(
                visit_unlock_pool, k=num_visit_unlocks_in_shop
            )
            self.shop_items.extend(chosen_unlocks)
            for chosen_unlock in chosen_unlocks:
                item_pool.remove(chosen_unlock)

        if settings.shop_keyblades:
            keyblade_item_pool = [
                i for i in item_pool if i.ItemType == itemType.KEYBLADE
            ]
            self.shop_items.extend(keyblade_item_pool)
            for keyblade_item in keyblade_item_pool:
                item_pool.remove(keyblade_item)
        return item_pool

    def initial_ability_pool(self, settings: RandomizerSettings) -> list[KH2Item]:
        """Builds the initial ability pool, based on settings."""
        modifier = SeedModifier.ability_list_modifier(settings.ability_pool_option)
        ability_pool: list[KH2Item] = modifier(
            Items.getActionAbilityList(),
            Items.getSupportAbilityList() + Items.getKeybladeAbilityList() + Items.getLevelAbilityList(),
        )

        # Remove any starting abilities from the pool
        # (make sure to only remove one copy for each starting item)
        remove_abilities = []
        for starting_item_id in settings.starting_inventory_ids:
            for pool_ability in ability_pool:
                if pool_ability.Id == starting_item_id:
                    remove_abilities.append(pool_ability)
                    break
        for remove_ability in remove_abilities:
            ability_pool.remove(remove_ability)

        return ability_pool

    def partition_locations(
        self, settings: RandomizerSettings
    ) -> tuple[list[KH2Location], list[KH2Location]]:
        """Splits all locations into lists of valid and invalid locations, based on settings."""
        all_locations = self.master_locations.all_locations()
        self.augment_invalid_checks(all_locations, settings)

        # If not "statsanity", even disabled locations should get stat bonuses
        if not settings.statSanity:
            self.assign_stat_bonuses(all_locations)

        def invalid_checker(location: KH2Location) -> bool:
            types = location.LocationTypes
            check_list = [
                locationType.OCCups,
                locationType.OCParadoxCup,
                locationType.CoR,
                locationType.TTR,
            ]
            invalid = any(
                disabled_type in types for disabled_type in settings.disabledLocations
            )
            if any(loc_type in check_list for loc_type in types):
                invalid = not any(
                    loc_type in settings.enabledLocations and loc_type in check_list
                    for loc_type in types
                )
            return invalid

        def remove_popupchecker(location: KH2Location) -> bool:
            if not settings.remove_popups:
                return False
            relevant_categories = [
                locationCategory.POPUP,
                locationCategory.DOUBLEBONUS,
                locationCategory.HYBRIDBONUS,
                locationCategory.ITEMBONUS,
                locationCategory.STATBONUS,
            ]
            superboss_types = [
                locationType.AS,
                locationType.DataOrg,
                locationType.LW,
                locationType.Sephi,
            ]
            category = location.LocationCategory
            if category in relevant_categories and not any(
                item in location.LocationTypes for item in superboss_types
            ):
                return True
            else:
                return False

        def no_final_form(location: KH2Location) -> bool:
            if not settings.disable_final_form:
                return False
            return location.LocationCategory is locationCategory.FINALLEVEL

        valid_locations: list[KH2Location] = []
        invalid_locations: list[KH2Location] = []
        excluded_levels = settings.excluded_levels()
        for loc in all_locations:
            if (
                no_final_form(loc)
                or invalid_checker(loc)
                or remove_popupchecker(loc)
                or (
                    loc.LocationCategory is locationCategory.LEVEL
                    and loc.LocationId in excluded_levels
                )
            ):
                invalid_locations.append(loc)
            else:
                valid_locations.append(loc)

        if len(all_locations) != (len(invalid_locations) + len(valid_locations)):
            raise GeneratorException(
                f"Separating valid {len(valid_locations)} and invalid {len(invalid_locations)} locations removed locations from existence (total {len(all_locations)})"
            )

        return valid_locations, invalid_locations

    @staticmethod
    def split_vanilla_abilities(
        ability_pool: list[KH2Item], vanilla_item_ids: list[int]
    ) -> tuple[list[KH2Item], list[KH2Item]]:
        """Splits the ability pool into lists of vanilla and randomizable abilities."""
        vanilla_item_ids_copy = copy.deepcopy(vanilla_item_ids)
        vanilla_abilities = []
        randomizable_abilities = []
        for item in ability_pool:
            if item.Id in vanilla_item_ids_copy:
                vanilla_item_ids_copy.remove(item.Id)
                vanilla_abilities.append(item)
            else:
                randomizable_abilities.append(item)
        return vanilla_abilities, randomizable_abilities

    def assign_sora_items(self, settings: RandomizerSettings):
        """Assigns items to locations for Sora."""
        self.apply_starting_items(settings)

        item_pool = self.initial_item_pool(settings)
        ability_pool = self.initial_ability_pool(settings)

        # Items in the shop that don't affect the item pool.
        if settings.shop_elixirs:
            self.shop_items.append(KH2Item(consumable.Elixir))
            self.shop_items.append(KH2Item(consumable.Megalixir))
        if settings.shop_recoveries:
            self.shop_items.append(KH2Item(consumable.DriveRecovery))
            self.shop_items.append(KH2Item(consumable.HighDriveRecovery))
        if settings.shop_boosts:
            self.shop_items.append(KH2Item(consumable.PowerBoost))
            self.shop_items.append(KH2Item(consumable.MagicBoost))
            self.shop_items.append(KH2Item(consumable.DefenseBoost))
            self.shop_items.append(KH2Item(consumable.ApBoost))

        valid_locations, invalid_locations = self.partition_locations(settings)

        locations_with_vanilla_items = [
            l
            for l in invalid_locations
            if len(l.VanillaItems) > 0
            and any(item in l.LocationTypes for item in settings.vanillaLocations)
        ]
        vanilla_item_ids = []
        for loc_with_vanilla in locations_with_vanilla_items:
            vanilla_item_ids.extend(loc_with_vanilla.VanillaItems)

        if len(vanilla_item_ids) > 0 and settings.chainLogic:
            raise GeneratorException(
                "Can't use vanilla worlds with chain logic for now. Sorry"
            )

        self.num_valid_locations = (
            len(valid_locations)
            + (
                len(
                    [
                        loc
                        for loc in valid_locations
                        if loc.LocationCategory
                        in [locationCategory.DOUBLEBONUS, locationCategory.HYBRIDBONUS]
                    ]
                )
                if settings.statSanity
                else 0
            )
            + (
                len(
                    [
                        loc
                        for loc in valid_locations
                        if loc.LocationCategory in [locationCategory.DOUBLEBONUS]
                    ]
                )
                if settings.statSanity
                else 0
            )
        )
        self.num_available_items = (
            len(ability_pool)
            + len(item_pool) 
            + (settings.max_objectives_available if settings.objective_rando else 0) 
            + (settings.max_emblems_available if settings.emblems else 0)
            - (sum([len(l.VanillaItems) for l in locations_with_vanilla_items]))
        )

        if self.progress_bar_vis:
            return

        # actual item assignment code starts here
        
        # give synth recipes their ingredients (if they are in valid locations)
        for location in valid_locations:
            if locationType.SYNTH in location.LocationTypes:
                # assign a recipe to this item
                items: list[KH2Item] = random.sample(
                    Items.getSynthRequirementsList(), k=random.randint(1, 3)
                )
                requirements = [
                    SynthRequirement(synth_item=item, amount=random.randint(1, 3))
                    for item in items
                ]
                recipe = SynthesisRecipe(
                    location=location,
                    unlock_rank=1 if location.LocationId < 15 else 2,
                    requirements=requirements,
                )
                self.synthesis_recipes.append(recipe)


        vanilla_abilities, randomizable_abilities = self.split_vanilla_abilities(
            ability_pool, vanilla_item_ids
        )
        valid_junk = self.get_n_junk(
            settings, num_junk_items=self.num_valid_locations - self.num_available_items
        )

        self.assign_keyblade_abilities(settings, randomizable_abilities, item_pool)
        item_pool.extend(vanilla_abilities)
        item_pool.extend(randomizable_abilities)
        item_pool.extend(valid_junk)
        if settings.emblems:
            item_pool.extend([Items.emblemItem() for _ in range(settings.max_emblems_available)])
        random.shuffle(item_pool)

        # vanilla location item assignment
        for loc_with_vanilla in locations_with_vanilla_items:
            # find item in item list
            for vanilla_item_id in loc_with_vanilla.VanillaItems:
                vanilla_item = next(
                    (item for item in item_pool if item.Id == vanilla_item_id), None
                )
                if vanilla_item is None:
                    # if we don't have the item, it means that we started with the item, or it was randomized away
                    continue
                if True: #vanilla_item.ItemType not in loc_with_vanilla.InvalidChecks:
                    item_pool.remove(vanilla_item)
                    if self.assign_item(loc_with_vanilla, vanilla_item, override_invalid=True):
                        invalid_locations.remove(loc_with_vanilla)

        # assign items that have very restricted locations (boss depths, yeet, etc.)
        self.assign_plando_like_items(settings, item_pool, valid_locations)

        # check for objective rando
        if settings.objective_rando:
            # get the objective pool
            objective_pool = settings.available_objectives
            # filter down the pool based on what valid locations there are
            valid_location_descriptions = [v.Description for v in valid_locations]
            objective_pool = [o for o in objective_pool if o.Location in valid_location_descriptions]
            # remove duplicates for AS/Data split
            if locationType.AS in settings.enabledLocations and locationType.DataOrg in settings.enabledLocations:
                if settings.as_data_split:
                    # remove duplicates for ASs
                    objective_pool = [o for o in objective_pool if "Defeat AS" not in o.Name]
                else:
                    # remove duplicates for Datas
                    objective_pool = [o for o in objective_pool if not any(t in o.Name for t in ["Data Vexen","Data Zexion","Data Larxene","Data Lexaeus","Data Marluxia"])]
            if settings.objective_pool_type==ObjectivePoolOption.HITLIST.name:
                # remove all but one of the form objectives
                form_objectives = [o for o in objective_pool if "Level 7" in o.Name]
                form_choice = []
                if len(form_objectives)>0:
                    form_choice = [random.choice(form_objectives)]

                objective_pool = [o for o in objective_pool if "Level 7" not in o.Name] + form_choice

            if len(objective_pool) < settings.max_objectives_available:
                raise SettingsException(f"Not enough objective locations ({len(objective_pool)}) available to allow the max number of objectives ({settings.max_objectives_available}) to be placed.")

            # remove duplicate form levels of the same level just to hope we don't have too many objectives on forms
            non_form_objectives = [o for o in objective_pool if "Level" not in o.Name]
            form_3_objectives = [o for o in objective_pool if "Level 3" in o.Name]
            form_5_objectives = [o for o in objective_pool if "Level 5" in o.Name]
            form_7_objectives = [o for o in objective_pool if "Level 7" in o.Name]
            random.shuffle(form_3_objectives)
            random.shuffle(form_5_objectives)
            random.shuffle(form_7_objectives)
            all_form_objectives = form_3_objectives + form_5_objectives + form_7_objectives
            if len(all_form_objectives) > 3:
                all_form_objectives[::3] = form_3_objectives
                all_form_objectives[1::3] = form_5_objectives
                all_form_objectives[2::3] = form_7_objectives
            # if the pool isn't going to be big enough, we need to fill with form levels
            if settings.max_objectives_available > len(non_form_objectives):
                non_form_objectives.extend(all_form_objectives[0:(settings.max_objectives_available-len(non_form_objectives))])
            else:
                # pick first 3
                non_form_objectives.extend(all_form_objectives[0:min(3, len(all_form_objectives))])
            objective_pool = non_form_objectives
                

            # pick a number of objectives
            self.objectives = random.sample(objective_pool,k=settings.max_objectives_available)
            picked_objectives_location_names = [o.Location for o in self.objectives]
            # plando completion marks onto the selected objectives
            objective_locations = [v for v in valid_locations if v.Description in picked_objectives_location_names]
            for o in objective_locations:
                locations_to_remove = self.randomly_assign_single_item(Items.objectiveItem(),[o])
                for l in locations_to_remove:
                    valid_locations.remove(l)
            

        if settings.chainLogic:
            self.assign_chain_logic(settings, item_pool, valid_locations)    
        else:
            # create some space for random assignment by doing a forward-pass assignment
            self.create_available_location_space(settings, item_pool, valid_locations)                

        self.randomly_assign_items(item_pool, valid_locations)

        # move remaining items to invalid locations for junk item assignment
        invalid_locations.extend(valid_locations)
        self.assign_junk_locations(settings, invalid_locations)

    def assign_chain_logic(self, settings : RandomizerSettings, item_pool, valid_locations):
        def contains(bigger_list,smaller_list):
            return not (Counter(smaller_list)-Counter(bigger_list))
        def missing_items(bigger_list,smaller_list):
            return list((Counter(smaller_list)-Counter(bigger_list)).elements())
        
        # this is checking for access into CoR, we'll need to simulate having the necessary movement
        simulated_growth = [growth.HighJump1.id,growth.HighJump2.id,growth.HighJump3.id,
                                                 growth.QuickRun1.id,growth.QuickRun2.id,growth.QuickRun3.id,
                                                 growth.AerialDodge1.id,growth.AerialDodge2.id,growth.AerialDodge3.id,
                                                 growth.Glide1.id,growth.Glide2.id,growth.Glide3.id,
                                                 ]
        
        shop_item_ids = [i.Id for i in self.shop_items]

        # determine available unlocks for
        #   1) regular/reverse
        #   2) keyblade locking on/off
        #   3) Anything already placed
        #       a) specifically visit locks and proofs from the "very restricted" pool
        #   4) Determine the existing dependency graph (we will use this to make sure we don't overlap two unlocks in the chain)
        from Module.seedEvaluation import LocationInformedSeedValidator
        validator = LocationInformedSeedValidator()
        validator.prep_requirements_list(settings, self)
        item_locking_ids_per_world = validator.generate_locking_item_ids()

        item_depth = 0
        min_item_depth = settings.chainLogicMinLength
        max_item_depth = settings.chainLogicMaxLength

        acquired_items,accessible_locations = self.get_accessible_locations(valid_locations,validator,simulated_growth + shop_item_ids)
        acquired_items = acquired_items + self.starting_item_ids + simulated_growth + shop_item_ids
        last_unlocked_sphere = accessible_locations

        world_names_to_unlock = [w for w in item_locking_ids_per_world.keys() for _ in item_locking_ids_per_world[w]]
        random.shuffle(world_names_to_unlock)
        world_names_to_unlock = [w for w in world_names_to_unlock if w in settings.enabledLocations]

        proof_nonexistence_assignment = self.assignment_for_item_id(proof.ProofOfNonexistence.id)
        proof_of_nonexistence_last_unlock = None

        second_pass_worlds = []

        if proof_nonexistence_assignment is not None:
            # find the world that has the proof, and how many items it takes to get it
            proof_world = list(set(proof_nonexistence_assignment.location.LocationTypes).intersection(item_locking_ids_per_world.keys()))[0]
            # figure out how far into the world you need to go
            world_progression = item_locking_ids_per_world[proof_world]
            items_needed = []
            for w in world_progression:
                if validator.is_location_available(self.starting_item_ids + items_needed, proof_nonexistence_assignment.location):
                    # uh, bad chain
                    raise GeneratorException("Chain logic couldn't make a seed that ended on proof of nonexistence. Try changing proof depths.")
                proof_of_nonexistence_last_unlock = missing_items(items_needed,w)
                items_needed.extend(proof_of_nonexistence_last_unlock)

            # confirmed location, check last instance of this world and move it last
            last_index = [i for i in range(len(world_names_to_unlock)) if world_names_to_unlock[i] == proof_world][-1]
            del world_names_to_unlock[last_index]
            second_pass_worlds = [proof_world]     

        history_of_items = []
        # print("Starting chain....")
        while (item_depth < min_item_depth or (item_depth < max_item_depth and random.random() < 0.75)): # 25% chance of breaking early
            # dummy_do_not_use = input("...")
            # print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            # print(f"Depth {item_depth}/({min_item_depth}-{max_item_depth})")
            # print("--iter")
            # pick a world to unlock checks from
            if world_names_to_unlock is None:
                break
            if len(world_names_to_unlock) == 0:
                break
            chosen_world = world_names_to_unlock[0]
            chosen_checks = item_locking_ids_per_world[chosen_world][0]
            # print(f"chosen checks {chosen_checks}")
            def pop_world_from_list():
                nonlocal world_names_to_unlock
                nonlocal second_pass_worlds
                world_names_to_unlock.pop(0)
                if len(world_names_to_unlock)==0:
                    world_names_to_unlock = second_pass_worlds
                    second_pass_worlds = None
                item_locking_ids_per_world[chosen_world].pop(0)
            if contains(acquired_items,chosen_checks):
                # print(f"chosen world {chosen_world}")
                # print(f"--Already have these items {chosen_checks}")
                pop_world_from_list()
                continue
            chosen_checks = missing_items(acquired_items,chosen_checks)
            sphere_1 = [loc for loc in valid_locations if locationType.Level not in loc.LocationTypes and validator.is_location_available(acquired_items + chosen_checks,loc) and loc not in accessible_locations]
            if settings.extended_placement_logic:
                sphere_1 = [loc for loc in sphere_1 if loc.Description!=hb.CheckLocation.DataDemyxApBoost]
            
            # print("*******************")
            # print(acquired_items)
            # print(chosen_checks)
            # print([s.Description for s in sphere_1])
            # print("*******************")
            if len(sphere_1) and len(last_unlocked_sphere) >= len(chosen_checks):
                pop_world_from_list()
                # assign new items
                history_of_items.append(chosen_checks)
                for i in chosen_checks:
                    # print(f"placing {i}")
                    i_data = next((it for it in item_pool if it.Id == i), None)
                    if i_data is not None:
                        # print(f"Assigning item {i_data} in one of these locations {last_unlocked_sphere}")
                        # assign this item somewhere
                        try:
                            locations_to_remove = self.randomly_assign_single_item(i_data,last_unlocked_sphere)
                        except Exception:
                            raise GeneratorException(f"Tried to assign {i_data} to {last_unlocked_sphere} and failed")
                        for l in locations_to_remove:
                            valid_locations.remove(l)
                        item_pool.remove(i_data)
                    else:
                        output_str = f"Uh....chain logic couldn't find an item with the id {i} from the unlock list. This shouldn't happen. History {history_of_items}"
                        print(output_str)
                        raise GeneratorException(output_str)
                acquired_items,accessible_locations = self.get_accessible_locations(valid_locations,validator, simulated_growth + shop_item_ids)
                acquired_items = acquired_items + self.starting_item_ids + simulated_growth + shop_item_ids
                last_unlocked_sphere = sphere_1
                # print(last_unlocked_sphere)
                item_depth+=1
            else:
                if chosen_world not in settings.enabledLocations:
                    # print(f"{chosen_world} checks are disabled")
                    pop_world_from_list()
                elif len(chosen_checks)==1 and chosen_checks[0] in [k.id for k in keyblade.get_locking_keyblades()] and settings.keyblades_unlock_chests:
                    pop_world_from_list()
                else:
                    # print(chosen_world)
                    # print(f"{len(sphere_1)} and {len(last_unlocked_sphere) }>={len(chosen_checks)} {chosen_checks}")
                    # print("--This item doesn't unlock enough right now, trying again later")
                    world_names_to_unlock.pop(0)
                    if second_pass_worlds is not None:
                        second_pass_worlds.insert(0,chosen_world)              

        if item_depth < min_item_depth:
            raise GeneratorException("Couldn't generate a chain long enough, are enough worlds on?...")

        i_data = next((it for it in item_pool if it.Id == proof.ProofOfNonexistence.id), None)
        if i_data is not None:
            # assign this item somewhere
            locations_to_remove = self.randomly_assign_single_item(i_data,last_unlocked_sphere)
            for l in locations_to_remove:
                valid_locations.remove(l)
            item_pool.remove(i_data)
        else:
            # make sure we have access to nonexistence
            acquired_items,accessible_locations = self.get_accessible_locations(valid_locations,validator, simulated_growth)
            acquired_items = acquired_items + self.starting_item_ids + simulated_growth
            if proof.ProofOfNonexistence.id not in acquired_items:
                raise GeneratorException("Couldn't access proof of nonexistence in chain logic")
            
    def create_available_location_space(self, settings, item_pool, valid_locations):
        # (a)determine sphere 0
        #    if sphere 0 is big enough
        #       return
        #    if sphere 0 is too small
        #    determine available unlocks for available checks
        #    place that somewhere that's available
        #    loop again        
        from Module.seedEvaluation import LocationInformedSeedValidator
        validator = LocationInformedSeedValidator()
        validator.prep_requirements_list(settings, self)

        shop_item_ids = [i.Id for i in self.shop_items]

        sphere_0_check = 30
        num_available_locations_needed_to_allow_random_assignment = 125 # TODO(zak) this is arbitrary and may need tuning
        while sphere_0_check > 0:
            # calculate the available checks
            _, sphere_0 = self.get_accessible_locations(valid_locations, validator, shop_item_ids)
            if len(sphere_0) < num_available_locations_needed_to_allow_random_assignment:
                # pick a locking item we don't have, assign it somewhere valid, repeat
                unlocks = [] + [s.id for s in storyunlock.all_story_unlocks()] + [k.id for k in keyblade.get_locking_keyblades()]
                random.shuffle(unlocks)
                for u in unlocks:
                    i_data = next((it for it in item_pool if it.Id == u), None)
                    if i_data is not None:
                        _, sphere_1 = self.get_accessible_locations(valid_locations, validator, shop_item_ids + [i_data.Id])
                        if len(sphere_1) > len(sphere_0):
                            # assign this item somewhere
                            locations_to_remove = self.randomly_assign_single_item(i_data,sphere_0)
                            for l in locations_to_remove:
                                valid_locations.remove(l)
                            item_pool.remove(i_data)
                            break
                sphere_0_check-=1 # make sure this doesn't ever cause a softlock again
            else:
                sphere_0_check = 0

    def get_accessible_locations(self, valid_locations, validator, aux_items = None):
        if aux_items is None:
            aux_items = []
        acquired_item_locations = []
        acquired_items = []
        sphere_0 = []
        found_new_item = True
        # check if any of the already assigned locations are valid right now, and if so, add their item to current inventory
        while found_new_item:
            found_new_item = False
            # get unassigned locations that are available
            sphere_0 = [loc for loc in valid_locations if validator.is_location_available(self.starting_item_ids + acquired_items + aux_items,loc)]
            # get already assigned items from available locations
            for assignment in self.assignments:
                if assignment.location.LocationCategory is not locationCategory.WEAPONSLOT and assignment.location not in acquired_item_locations and validator.is_location_available(self.starting_item_ids + acquired_items + aux_items, assignment.location):
                    if assignment.location.Description != stt.CheckLocation.StruggleWinnerChampionBelt:
                        acquired_item_locations.append(assignment.location)
                        acquired_items.extend([i.Id for i in assignment.items()])
                        found_new_item = True
        return acquired_items,sphere_0

    def assign_plando_like_items(self, settings, item_pool, valid_locations):
        restricted_reports = self.report_depths.very_restricted_locations
        restricted_proofs = self.proof_depths.very_restricted_locations
        restricted_promise_charm = self.promise_charm_depths.very_restricted_locations
        restricted_story = self.story_depths.very_restricted_locations

        if self.yeet_the_bear:
            # Manually assign to Starry Hill
            yeet_locations = _find_locations(
                haw.yeet_the_bear_location_names(), valid_locations
            )
            if len(yeet_locations) > 0:
                yeet_location = random.choice(yeet_locations)
                proof_item = next(
                        key for key in item_pool if key.Id == proof.ProofOfNonexistence.id
                    )
                if self.assign_item(yeet_location, proof_item):
                    valid_locations.remove(yeet_location)
                    item_pool.remove(proof_item)
            else:
                raise CantAssignItemException(
                    "None of the Starry Hill locations are available for Yeet the Bear. Is HAW not randomized?"
                )
            
        if restricted_proofs:
            remaining_proofs = [i for i in item_pool if i.ItemType in proof.proof_item_types()]
            # pick N valid locations for these items
            valid_for_proofs = [loc for loc in valid_locations if self.proof_depths.is_valid(loc)]

            # Check if both AS's and datas are enabled, and remove AS locations if so
            # We do this mainly to prevent the AS and data fight in the same world from both having proofs.
            if (
                locationType.AS in settings.enabledLocations
                and locationType.DataOrg in settings.enabledLocations
            ):
                valid_for_proofs = [loc for loc in valid_for_proofs if locationType.AS not in loc.LocationTypes]

            good_choices = False
            while not good_choices:
                good_choices = True
                chosen_locations = random.sample(valid_for_proofs,k=len(remaining_proofs))
                for index,c in enumerate(chosen_locations):
                    # check that each proof is valid for the location (i.e. no connection on Terra)
                    if remaining_proofs[index].ItemType in c.InvalidChecks:
                        good_choices = False                    

            for index,c in enumerate(chosen_locations):
                item_pool.remove(remaining_proofs[index])
                if self.assign_item(c, remaining_proofs[index]):
                    valid_locations.remove(c)
        if restricted_promise_charm:
            promise_charm_item_list = [i for i in item_pool if i.ItemType is misc.PromiseCharm.type]
            # pick N valid locations for these items
            valid_for_promise_charm = [loc for loc in valid_locations if self.promise_charm_depths.is_valid(loc)]
            chosen_locations = random.sample(valid_for_promise_charm,k=len(promise_charm_item_list))
            for index,c in enumerate(chosen_locations):
                item_pool.remove(promise_charm_item_list[index])
                if self.assign_item(c, promise_charm_item_list[index]):
                    valid_locations.remove(c)
        if restricted_story:
            remaining_unlocks = [i for i in item_pool if i.ItemType is itemType.STORYUNLOCK]
            # pick N valid locations for these items
            valid_for_unlocks = [loc for loc in valid_locations if self.story_depths.is_valid(loc)]
            number_of_choices = min(len(remaining_unlocks),len(valid_for_unlocks))
            chosen_locations = random.sample(valid_for_unlocks,k=number_of_choices)
            for index,c in enumerate(chosen_locations):
                item_pool.remove(remaining_unlocks[index])
                if self.assign_item(c, remaining_unlocks[index]):
                    valid_locations.remove(c)
        if restricted_reports:
            remaining_reports = [i for i in item_pool if i.ItemType is itemType.REPORT]
            # pick N valid locations for these items
            valid_for_reports = [loc for loc in valid_locations if self.report_depths.is_valid(loc)]
            number_of_choices = min(len(remaining_reports),len(valid_for_reports))
            chosen_locations = random.sample(valid_for_reports,k=number_of_choices)
            for index,c in enumerate(chosen_locations):
                item_pool.remove(remaining_reports[index])
                if self.assign_item(c, remaining_reports[index]):
                    valid_locations.remove(c)

    def randomly_assign_single_item(self, item, location_pool):
        locations_to_remove = []
        if len(location_pool) == 0:
            raise CantAssignItemException(f"Ran out of locations to assign items {item}")

        weights = self.compute_location_weights(item, location_pool)

        count = 0
        while True:
            count += 1
            if len(weights) == 0:
                raise CantAssignItemException(
                    f"Ran out of locations to assign items to."
                )
            if sum(weights) == 0:
                raise CantAssignItemException(
                    f"Somehow, can't assign an item because there are no valid locations for {item.Name}."
                )

            random_location: KH2Location = random.choices(location_pool, weights)[
                0
            ]
            if item.ItemType not in random_location.InvalidChecks:
                if self.assign_item(random_location, item):
                    location_pool.remove(random_location)
                    locations_to_remove.append(random_location)

                    struggle_pair = self._maybe_assign_struggle_pair(
                        random_location, item, location_pool
                    )
                    if struggle_pair is not None:
                        location_pool.remove(struggle_pair)
                        locations_to_remove.append(struggle_pair)
                break
            if count == 100:
                raise CantAssignItemException(
                    f"Trying to assign {item} and failed 100 times in {len([i for i in location_pool if i.LocationCategory==locationCategory.POPUP])} popups left out of {len(location_pool)}"
                )
        return locations_to_remove


    def randomly_assign_items(self, item_pool, valid_locations):
        for item in item_pool:
            if len(valid_locations) == 0:
                raise CantAssignItemException(f"Ran out of locations to assign items")

            weights = self.compute_location_weights(item, valid_locations)

            count = 0
            while True:
                count += 1
                if len(weights) == 0:
                    raise CantAssignItemException(
                        f"Ran out of locations to assign items to."
                    )
                if sum(weights) == 0:
                    raise CantAssignItemException(
                        f"Somehow, can't assign an item because there are no valid locations for {item.Name}."
                    )

                random_location: KH2Location = random.choices(valid_locations, weights)[
                    0
                ]
                if item.ItemType not in random_location.InvalidChecks:
                    if self.assign_item(random_location, item):
                        valid_locations.remove(random_location)

                        struggle_pair = self._maybe_assign_struggle_pair(
                            random_location, item, valid_locations
                        )
                        if struggle_pair is not None:
                            valid_locations.remove(struggle_pair)
                    break
                if count == 100:
                    raise CantAssignItemException(
                        f"Trying to assign {item} and failed 100 times in {len([i for i in valid_locations if i.LocationCategory==locationCategory.POPUP])} popups left out of {len(valid_locations)}"
                    )


    def compute_location_weights(
        self, item: KH2Item, location_pool: list[KH2Location]
    ) -> list[int]:
        loc_weights = self.location_weights
        result = [
            loc_weights.get_weight(item.ItemType, loc) for loc in location_pool
        ]
        return result

    def assign_junk_locations(
        self, settings: RandomizerSettings, locations: list[KH2Location]
    ):
        """Assign the rest of the locations with "junk"."""
        junk_items = [
            item
            for item in Items.getJunkList(betterJunk=False)
            if item.Id in settings.junk_pool
        ]

        # Assign the same item to both struggle winner and loser if those locations are still available
        struggle_winner = _find_location(
            stt.CheckLocation.StruggleWinnerChampionBelt, locations
        )
        struggle_loser = _find_location(stt.CheckLocation.StruggleLoserMedal, locations)
        if struggle_winner is not None and struggle_loser is not None:
            junk_item = random.choice(junk_items)
            self.assign_item(struggle_winner, junk_item)
            self.assign_item(struggle_loser, junk_item)
            locations.remove(struggle_winner)
            locations.remove(struggle_loser)
        elif struggle_winner is not None and struggle_loser is None:
            raise GeneratorException(
                "Attempting to assign junk to struggle winner but loser already has an item"
            )
        elif struggle_loser is not None and struggle_winner is None:
            raise GeneratorException(
                "Attempting to assign junk to struggle loser but winner already has an item"
            )

        excluded_levels = settings.excluded_levels()
        for loc in locations:
            if loc.LocationCategory is not locationCategory.LEVEL or (
                loc.LocationCategory is locationCategory.LEVEL
                and loc.LocationId not in excluded_levels
            ):
                junk_item = random.choice(junk_items)
                # assign another junk item if that location needs another item
                if not self.assign_item(loc, junk_item):
                    junk_item = random.choice(junk_items)
                    self.assign_item(loc, junk_item)
            else:
                self.assign_item(loc, Items.getNullItem())

    @staticmethod
    def get_n_junk(settings: RandomizerSettings, num_junk_items: int) -> list[KH2Item]:
        """Returns a given number of "junk" items."""
        all_junk_items = [
            item
            for item in Items.getJunkList(betterJunk=False)
            if item.Id in settings.junk_pool
        ]
        return random.choices(all_junk_items, k=num_junk_items)

    def augment_invalid_checks(
        self, locations: list[KH2Location], settings: RandomizerSettings
    ):
        """Add invalid check types to locations."""
        for loc in locations:
            if loc.LocationCategory in [
                locationCategory.POPUP,
                locationCategory.CREATION,
            ]:
                loc.InvalidChecks.append(itemType.GROWTH_ABILITY)
                loc.InvalidChecks.append(itemType.ACTION_ABILITY)
                loc.InvalidChecks.append(itemType.SUPPORT_ABILITY)
                loc.InvalidChecks.append(itemType.GAUGE)
            if (
                locationType.STT in loc.LocationTypes
                and loc.LocationCategory != locationCategory.STATBONUS
            ):
                loc.InvalidChecks.append(itemType.GAUGE)
            if locationType.Critical in loc.LocationTypes:
                loc.InvalidChecks.append(itemType.GAUGE)

            # handle the easy depth options, restricted versions are handled specially
            if not self.report_depths.very_restricted_locations:
                if not self.report_depths.is_valid(loc):
                    loc.InvalidChecks.append(itemType.REPORT)

            if not self.proof_depths.very_restricted_locations:
                if not self.proof_depths.is_valid(loc):
                    loc.InvalidChecks.extend(proof.proof_item_types())

            if not self.story_depths.very_restricted_locations:
                if not self.story_depths.is_valid(loc):
                    loc.InvalidChecks.append(itemType.STORYUNLOCK)

            if not self.promise_charm_depths.very_restricted_locations:
                if not self.promise_charm_depths.is_valid(loc):
                    loc.InvalidChecks.append(itemType.PROMISE_CHARM)

            ''' Commenting this code out because new plando logic should account for this
            # if both reports and proofs are very restricted (only in 13 locations)
            # add extra proof restrictions to allow reports to be assigned
            if self.report_depths.is_valid(loc) and self.proof_depths.is_valid(loc):
                if (
                    self.report_depths.very_restricted_locations
                    and self.proof_depths.very_restricted_locations
                ):
                    loc.InvalidChecks.extend(proof.proof_item_types())

            if (
                self.proof_depths.is_valid(loc)
                and self.proof_depths.very_restricted_locations
            ):
                # Check if both AS's and datas are enabled, and if this is an AS location, disable that.
                # We do this mainly to prevent the AS and data fight in the same world from both having proofs.
                if (
                    locationType.AS in settings.enabledLocations
                    and locationType.DataOrg in settings.enabledLocations
                ):
                    if locationType.AS in loc.LocationTypes:
                        loc.InvalidChecks.extend(proof.proof_item_types())

            # Only allow Proof of Nonexistence on one of the yeet the bear check locations
            # (this may override a proof depth option but player chose to put the proof here)
            if self.yeet_the_bear:
                if loc.name() in haw.yeet_the_bear_location_names():
                    while itemType.PROOF_OF_NONEXISTENCE in loc.InvalidChecks:
                        loc.InvalidChecks.remove(itemType.PROOF_OF_NONEXISTENCE)
                else:
                    loc.InvalidChecks.append(itemType.PROOF_OF_NONEXISTENCE)
            '''

    def assign_keyblade_abilities(
        self,
        settings: RandomizerSettings,
        ability_pool: list[KH2Item],
        item_pool: list[KH2Item],
    ):
        """Assign abilities to keyblades."""
        eligible_ids = set(
            settings.keyblade_support_abilities + settings.keyblade_action_abilities
        )

        # remove auto abilities from keyblades
        if settings.extended_placement_logic:
            eligible_ids.discard(ability.AutoValor.id)
            eligible_ids.discard(ability.AutoWisdom.id)
            eligible_ids.discard(ability.AutoMaster.id)
            eligible_ids.discard(ability.AutoFinal.id)
            eligible_ids.discard(ability.AutoLimitForm.id)

        eligible_abilities = [abil for abil in ability_pool if abil.Id in eligible_ids]
        nightmare_rarity_weights = {
            itemRarity.COMMON: 1,
            itemRarity.UNCOMMON: 2,
            itemRarity.RARE: 5,
            itemRarity.MYTHIC: 5,
        }

        keyblade_ids_to_exclude = [
            weaponslot.LocationId.KingdomKeyD,
            weaponslot.LocationId.AlphaWeapon,
            weaponslot.LocationId.OmegaWeapon,
            weaponslot.LocationId.KingdomKey,
        ]

        # assign all the abilities for keyblades
        for key in weaponslot.keyblade_slots():
            if (
                key.LocationId == weaponslot.LocationId.Pureblood
                and not settings.pureblood
            ):
                continue

            if len(eligible_abilities) == 0:
                raise GeneratorException(
                    "Keyblades: Not enough abilities are available to assign an ability to every keyblade"
                )

            if (
                settings.extended_placement_logic
                and key.LocationId not in keyblade_ids_to_exclude
            ):
                ability_weights = [
                    nightmare_rarity_weights[abil.Rarity] for abil in eligible_abilities
                ]
            else:
                ability_weights = [1 for _ in eligible_abilities]

            random_ability = random.choices(eligible_abilities, ability_weights)[0]
            self.assign_item(key, random_ability)
            ability_pool.remove(random_ability)
            eligible_abilities.remove(random_ability)

        # Assign draws to struggle weapons
        for weapon in weaponslot.struggle_weapon_slots():
            self.assign_item(weapon, KH2Item(ability.Draw))

    def assign_stat_bonuses(self, avail_locations: list[KH2Location]):
        """Assign all the stat items to bonuses for stats. Only used when "statsanity" is off."""
        stat_items = Items.getStatItems()
        double_stat = [
            loc
            for loc in avail_locations
            if loc.LocationCategory == locationCategory.DOUBLEBONUS
        ]
        single_stat = [
            loc
            for loc in avail_locations
            if loc.LocationCategory
            in [locationCategory.STATBONUS, locationCategory.HYBRIDBONUS]
        ]

        if len(double_stat) != 1:
            raise GeneratorException(
                f"Somehow have two locations with double stat gains {double_stat}"
            )

        # Select two different stats to put on Xemnas 1
        stat1 = random.choice(stat_items)
        stat_items.remove(stat1)
        stat2 = stat1
        while stat1 == stat2:
            stat2 = random.choice(stat_items)
        stat_items.remove(stat2)
        self.assign_item(double_stat[0], stat1)
        self.assign_item(double_stat[0], stat2)
        avail_locations.remove(double_stat[0])

        if len(single_stat) != len(stat_items):
            raise GeneratorException(
                f"The number of stat bonus locations {len(single_stat)} doesn't match remaining stat items {len(stat_items)}"
            )

        # Assign the rest
        for item in stat_items:
            loc = random.choice(single_stat)
            single_stat.remove(loc)
            if self.assign_item(loc, item):
                avail_locations.remove(loc)

        if len(single_stat) != 0:
            raise GeneratorException(
                f"Leftover stat locations were not assigned. Num remaining {len(single_stat)}"
            )

    def assign_item(
        self,
        location: KH2Location,
        item: KH2Item,
        party_member_assigned_items: Optional[list[ItemAssignment]] = None,
        override_invalid = False,
    ) -> bool:
        """
        Assigns the given item to the given location. Returns True if the location has all needed items assigned, or
        False if it is a location with two slots and the other slot still needs to be filled.
        """
        double_item = location.LocationCategory in [
            locationCategory.DOUBLEBONUS,
            locationCategory.HYBRIDBONUS,
        ]

        assigned_items = party_member_assigned_items
        if assigned_items is None:
            assigned_items = self.assignments

        if item.ItemType in location.InvalidChecks and not override_invalid:
            raise GeneratorException(
                f"Trying to assign {item} to {location} even though it's invalid."
            )

        assignment = next(
            filter(lambda a: a.location == location, assigned_items), None
        )
        if assignment is None:
            assigned_items.append(ItemAssignment(location, item))
            all_slots_filled = not double_item
        else:
            if assignment.item is None:
                raise GeneratorException(
                    f"Somehow assigned no item to a location {assignment}"
                )
            if not double_item:
                raise GeneratorException(
                    f"Assigning a second item to a location that can't have one {assignment}"
                )
            if assignment.item2 is not None:
                raise GeneratorException(
                    f"Assigning a third item to a location that already has two {assignment}"
                )
            assignment.item2 = item
            all_slots_filled = True

        return all_slots_filled

    def _maybe_assign_struggle_pair(
        self, loc: KH2Location, item: KH2Item, available_locations: list[KH2Location]
    ) -> Optional[KH2Location]:
        """
        If the assigned location is the struggle winner or loser, assigns the same item to the opposite.
        Returns the paired struggle location if the assigned location was either of the struggle winner or loser, or
        returns None otherwise.
        """
        if loc.name() == stt.CheckLocation.StruggleWinnerChampionBelt:
            opposite = _find_location(
                stt.CheckLocation.StruggleLoserMedal, available_locations
            )
        elif loc.name() == stt.CheckLocation.StruggleLoserMedal:
            opposite = _find_location(
                stt.CheckLocation.StruggleWinnerChampionBelt, available_locations
            )
        else:
            return None
        if opposite is None:
            raise GeneratorException("Tried assigning struggle reward, but failed")

        self.assign_item(opposite, item)
        return opposite

    def assignment_for_location(self, location_name: str) -> Optional[ItemAssignment]:
        return next(
            (a for a in self.assignments if a.location.name() == location_name), None
        )

    def assignment_for_item(self, item: InventoryItem) -> Optional[ItemAssignment]:
        return next((a for a in self.assignments if a.item.item == item), None)
    
    def assignment_for_item_id(self, item_id: int) -> Optional[ItemAssignment]:
        return next((a for a in self.assignments if a.item.item.id == item_id), None)



from Class.newLocationClass import KH2Location
from List.configDict import locationType
from List.location.graph import double_bonus, item_bonus


def goofy_bonuses() -> list[KH2Location]:
    return [
        double_bonus(21, "Barbossa", locationType.PR),
        item_bonus(59, "Grim Reaper 1", locationType.PR),
        item_bonus(31, "Hostile Program", locationType.SP),
        item_bonus(49, "Hyenas 1", locationType.PL),
        item_bonus(50, "Hyenas 2", locationType.PL),
        item_bonus(40, "Lock/Shock/Barrel", locationType.HT),
        item_bonus(19, "Oogie Boogie", locationType.HT),
        item_bonus(6, "Pete (Olympus)", locationType.OC),
        item_bonus(17, "Pete (Wharf)", locationType.DC),
        item_bonus(9, "Shan-Yu", locationType.LoD),
        item_bonus(10, "Storm Rider", locationType.LoD),
        item_bonus(12, "Beast", locationType.BC),
        item_bonus(39, "Interceptor Barrels", locationType.PR),
        item_bonus(46, "Treasure Room Heartless", locationType.Agrabah),
        item_bonus(66, "Zexion", [locationType.OC, locationType.AS]),
    ]



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Armor(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.ARMOR)


ElvenBandanna = Armor(67, "Elven Bandana")
DivineBandanna = Armor(68, "Divine Bandana")
PowerBand = Armor(69, "Power Band")
BusterBand = Armor(70, "Buster Band")
ProtectBelt = Armor(78, "Protect Belt")
GaiaBelt = Armor(79, "Gaia Belt")
CosmicBelt = Armor(111, "Cosmic Belt")
ShockCharm = Armor(132, "Shock Charm")
ShockCharmPlus = Armor(133, "Shock Charm+")
GrandRibbon = Armor(157, "Grand Ribbon")
FireBangle = Armor(173, "Fire Bangle")
FiraBangle = Armor(174, "Fira Bangle")
FiragaBangle = Armor(197, "Firaga Bangle")
FiragunBangle = Armor(284, "Firagun Bangle")
BlizzardArmlet = Armor(286, "Blizzard Armlet")
BlizzaraArmlet = Armor(287, "Blizzara Armlet")
BlizzagaArmlet = Armor(288, "Blizzaga Armlet")
BlizzagunArmlet = Armor(289, "Blizzagun Armlet")
ThunderTrinket = Armor(291, "Thunder Trinket")
ThundaraTrinket = Armor(292, "Thundara Trinket")
ThundagaTrinket = Armor(293, "Thundaga Trinket")
ThundagunTrinket = Armor(294, "Thundagun Trinket")
ShadowAnklet = Armor(296, "Shadow Anklet")
DarkAnklet = Armor(297, "Dark Anklet")
MidnightAnklet = Armor(298, "Midnight Anklet")
ChaosAnklet = Armor(299, "Chaos Anklet")
AbasChain = Armor(301, "Abas Chain")
AegisChain = Armor(302, "Aegis Chain")
Acrisius = Armor(303, "Acrisius")
Ribbon = Armor(304, "Ribbon")
ChampionBelt = Armor(305, "Champion Belt")
PetiteRibbon = Armor(306, "Petite Ribbon")
AcrisiusPlus = Armor(307, "Acrisius+")
CosmicChain = Armor(308, "Cosmic Chain")



import json
import os
import random
import string
import argparse

from collections import namedtuple

from Class.seedSettings import ExtraConfigurationData, SeedSettings
from Module import appconfig
from Module.RandomizerSettings import RandomizerSettings
from Module.generate import generateSeedCLI
from Module.seedshare import SharedSeed
from Module.version import LOCAL_UI_VERSION

# CLI arguments & docs
parser = argparse.ArgumentParser(
    description="A command line utility for generating KH2 Randomizer seed strings from a given preset"
)
parser.add_argument(
    "preset",
    help='preset from which to generate seed string (default: "League Summer 2025")',
    nargs="?",
    default="League Summer 2025",
)
args = parser.parse_args()

SeedInfo = namedtuple(
    "SeedInfo",
    ["seed_name", "requested_preset", "generator_string", "hash_icons", "spoiler_html"],
)

requested_preset = args.preset


def make_random_seed_from_preset_name(requested_type: str):
    # TODO: Search for the specific preset file instead of loading all of them
    # load preset data for all presets in folder
    preset_json = {}
    for preset_file_name in os.listdir(appconfig.PRESET_FOLDER):
        preset_name, extension = os.path.splitext(preset_file_name)
        if extension == ".json":
            with open(
                os.path.join(appconfig.PRESET_FOLDER, preset_file_name), "r"
            ) as presetData:
                try:
                    settings_json = json.load(presetData)
                    preset_json[preset_name] = settings_json
                except Exception:
                    print(
                        "Unable to load preset [{}], skipping".format(preset_file_name)
                    )

    # get seed name at random
    characters = string.ascii_letters + string.digits
    seedString = "".join(random.choice(characters) for i in range(30))
    makeSpoilerLog = False
    settings = SeedSettings()
    try:
        settings.apply_settings_json(preset_json[requested_type])
    except KeyError:
        print(
            '\033[31mNo preset file found with the name \033[1m"{}"\033[0m'.format(
                requested_type
            )
        )
        exit(128)

    shared_seed = SharedSeed(
        generator_version=LOCAL_UI_VERSION,
        seed_name=seedString,
        spoiler_log=makeSpoilerLog,
        settings_string=settings.settings_string(),
        tourney_gen=True,
    )
    shared_string_text = shared_seed.to_share_string()

    rando_settings = RandomizerSettings(
        seedString, makeSpoilerLog, LOCAL_UI_VERSION, settings, shared_string_text
    )

    extra_data = ExtraConfigurationData(
        platform="PC", tourney=True, custom_cosmetics_executables=[]
    )

    spoiler_log = generateSeedCLI(rando_settings, extra_data)

    return SeedInfo(
        seed_name=seedString,
        requested_preset=requested_type,
        generator_string=shared_string_text,
        hash_icons=rando_settings.seedHashIcons,
        spoiler_html=spoiler_log,
    )


if __name__ == "__main__":
    seed_info = make_random_seed_from_preset_name(requested_preset)
    print(seed_info.generator_string)
    print(seed_info.hash_icons)



import math

from Class.exceptions import SettingsException
from List.configDict import expCurve
from List.inventory import form
from List.inventory.form import DriveForm


def vanilla_form_exp(drive_form: DriveForm) -> dict[int, int]:
    if drive_form == form.ValorForm:
        return {1: 80, 2: 160, 3: 280, 4: 448, 5: 560, 6: 672, 7: 0}  # 2200 (520 + 1680) (~1:3.23)
    elif drive_form == form.WisdomForm:
        return {1: 12, 2: 24, 3: 48, 4: 76, 5: 133, 6: 157, 7: 0}  # 450 (84 + 366) (~1:4.36)
    elif drive_form == form.LimitForm:
        return {1: 3, 2: 6, 3: 12, 4: 19, 5: 23, 6: 27, 7: 0}  # 90 (21 + 69) (~1:3.28)
    elif drive_form == form.MasterForm:
        return {1: 40, 2: 80, 3: 140, 4: 224, 5: 448, 6: 668, 7: 0}  # 1600 (260 + 1340) (~1:5.15)
    elif drive_form == form.FinalForm:
        return {1: 12, 2: 24, 3: 48, 4: 76, 5: 133, 6: 157, 7: 0}  # 450 (84 + 366) (~1:4.36)


def midday_form_exp(drive_form: DriveForm) -> dict[int, int]:
    if drive_form == form.ValorForm:
        return {1: 92, 2: 184, 3: 324, 4: 266, 5: 334, 6: 400, 7: 0}  # 1600 (600 + 1000)
    elif drive_form == form.WisdomForm:
        return {1: 14, 2: 29, 3: 57, 4: 41, 5: 73, 6: 86, 7: 0}  # 300 (100 + 200)
    elif drive_form == form.LimitForm:
        return {1: 4, 2: 8, 3: 13, 4: 11, 5: 12, 6: 12, 7: 0}  # 60 (25 + 35)
    elif drive_form == form.MasterForm:
        return {1: 62, 2: 123, 3: 215, 4: 134, 5: 266, 6: 400, 7: 0}  # 1200 (400 + 800)
    elif drive_form == form.FinalForm:
        return {1: 14, 2: 29, 3: 57, 4: 41, 5: 73, 6: 86, 7: 0}  # 300 (100 + 200)


def dusk_form_exp(drive_form: DriveForm) -> dict[int, int]:
    if drive_form == form.ValorForm:
        return {1: 115, 2: 230, 3: 405, 4: 226, 5: 284, 6: 340, 7: 0}  # 1600 (750 + 850)
    elif drive_form == form.WisdomForm:
        return {1: 20, 2: 40, 3: 80, 4: 33, 5: 58, 6: 69, 7: 0}  # 300 (140 + 160)
    elif drive_form == form.LimitForm:
        return {1: 4, 2: 9, 3: 15, 4: 9, 5: 11, 6: 12, 7: 0}  # 60 (28 + 32)
    elif drive_form == form.MasterForm:
        return {1: 85, 2: 170, 3: 295, 4: 109, 5: 216, 6: 325, 7: 0}  # 1200 (550 + 650)
    elif drive_form == form.FinalForm:
        return {1: 20, 2: 40, 3: 80, 4: 33, 5: 58, 6: 69, 7: 0}  # 300 (140 + 160)


def vanilla_summon_exp() -> dict[int, int]:
    return {1: 6, 2: 16, 3: 25, 4: 42, 5: 63, 6: 98, 7: 0}  # 250 (57 + 193) (~1:3.38)


def midday_summon_exp() -> dict[int, int]:
    return {1: 8, 2: 19, 3: 48, 4: 25, 5: 42, 6: 58, 7: 0}  # 200 (75 + 125)


def dusk_summon_exp() -> dict[int, int]:
    return {1: 9, 2: 25, 3: 56, 4: 24, 5: 36, 6: 50, 7: 0}  # 200 (90 + 110)


def vanilla_sora_exp() -> list[int]:
    return [
        0,
        40,
        100,
        184,
        296,
        440,
        620,
        840,
        1128,
        1492,
        1940,
        2480,
        3120,
        3902,
        4838,
        5940,
        7260,
        8814,
        10618,
        12688,
        15088,
        17838,
        20949,
        24433,
        28302,
        32622,
        37407,
        42671,
        48485,
        54865,
        61886,
        69566,
        77984,
        87160,
        97177,
        108057,
        119887,
        132691,
        146560,
        161520,
        177666,
        195026,
        213699,
        233715,
        255177,
        278117,
        302642,
        328786,
        356660,
        386378,
        417978,
        450378,
        483578,
        517578,
        552378,
        587978,
        624378,
        661578,
        699578,
        738378,
        777978,
        818378,
        859578,
        901578,
        944378,
        987978,
        1032378,
        1077578,
        1123578,
        1170378,
        1217978,
        1266378,
        1315578,
        1365578,
        1416378,
        1467978,
        1520378,
        1573578,
        1627578,
        1682378,
        1737978,
        1794378,
        1851578,
        1909578,
        1968378,
        2027978,
        2088378,
        2149578,
        2211578,
        2274378,
        2337978,
        2402378,
        2467578,
        2533578,
        2600378,
        2667978,
        2736378,
        2805578,
        2875578,
        2875578
    ]


def create_diffs(experiences):
    diffs = []
    for i in range(len(experiences)-1):
        diffs.append(experiences[i+1]-experiences[i])
    return diffs


def get_base_exp_progression():
    vanilla = vanilla_sora_exp()[:-1]
    diffs = create_diffs(vanilla)
    diffs2 = create_diffs(diffs)
    diffs3 = create_diffs(diffs2)
    return diffs3


def create_new_values(second_deriv=20, first_deriv=40, early_offset=0, mid_offset=0, late_offset=0) -> list[int]:
    diffs3 = get_base_exp_progression()
    for i in range(25):
        diffs3[i] += early_offset

    for i in range(25, 48):
        diffs3[i] += mid_offset

    diffs3[48] += late_offset

    diffs2 = [second_deriv]
    for d in diffs3:
        diffs2.append(diffs2[-1] + d)
    diffs = [first_deriv]
    for d in diffs2:
        diffs.append(diffs[-1] + d)
    completed = [0]
    for d in diffs:
        completed.append(completed[-1] + d)
    completed.append(completed[-1])

    for i in range(len(completed) - 1):
        if completed[i + 1] < completed[i]:
            print("Something went wrong")
    return completed


def midday_sora_exp(adjusted: bool) -> list[int]:
    if adjusted:
        return create_new_values(20, 40, 2, -25, 400)
    else:
        return create_new_values(20, 40, 2, 2, -800)


def dusk_sora_exp(adjusted: bool) -> list[int]:
    if adjusted:
        return create_new_values(20, 40, 3, -50, 1200)
    else:
        return create_new_values(20, 40, 3, 3, -1100)


def get_sora_exp(max_level_checks: int, rate: float, curve: expCurve) -> list[int]:
    adjusted = max_level_checks == 50
    if curve == expCurve.DAWN:
        exp_list = vanilla_sora_exp()
    elif curve == expCurve.MIDDAY:
        exp_list = midday_sora_exp(adjusted)
    elif curve == expCurve.DUSK:
        exp_list = dusk_sora_exp(adjusted)
    else:
        raise SettingsException(f"Incorrect exp curve value {curve}")
    return [math.ceil(a / b) for a, b in zip(exp_list, [rate] * 100)]

# For now, not considering DAWN, MIDDAY or DUSK shenanigans
def get_companion_exp(rate: float) -> list[int]:
    exp_list = vanilla_sora_exp()
    return [math.ceil(a / b) for a, b in zip(exp_list, [rate] * 100)]


def get_form_exp(drive_form: DriveForm, rate: float, curve: expCurve) -> list[int]:
    if curve == expCurve.DAWN:
        exp_list = vanilla_form_exp(drive_form)
    elif curve == expCurve.MIDDAY:
        exp_list = midday_form_exp(drive_form)
    elif curve == expCurve.DUSK:
        exp_list = dusk_form_exp(drive_form)
    else:
        raise SettingsException(f"Incorrect exp curve value {curve}")
    return [math.ceil(a / b) for a, b in zip((exp_list[i] for i in range(1, 8)), [rate] * 7)]


def get_summon_exp(rate: float, curve: expCurve) -> list[int]:
    if curve == expCurve.DAWN:
        exp_list = vanilla_summon_exp()
    elif curve == expCurve.MIDDAY:
        exp_list = midday_summon_exp()
    elif curve == expCurve.DUSK:
        exp_list = dusk_summon_exp()
    else:
        raise SettingsException(f"Incorrect exp curve value {curve}")
    return [math.ceil(a / b) for a, b in zip((exp_list[i] for i in range(1, 8)), [rate] * 7)]



from enum import Enum

from List.configDict import locationType
from List.inventory import magic, keyblade, ability, misc, summon, report
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, hybrid_bonus, stat_bonus, item_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    Rampart = "Ramparts"
    RampartChests = "Ramparts Chests"
    PortRoyalTown = "PR Town"
    PortRoyalTownChests = "PR Town Chests"
    CaveMouth = "Cave Mouth"
    CaveMouthChests = "Cave Mouth Chests"
    IslaDeMuertaPopup = "Isla de Meurta Popup"
    BoatFight = "Boat Fight"
    BarrelsMinigame = "Barrels Minigame"
    PowderStore = "Powder Store"
    PowderStoreChests = "Powder Store Chests"
    MoonlightNook = "Moonlight Nook"
    MoonlightNookChests = "Moonlight Nook Chests"
    Barbossa = "Barbossa"
    GrimReaper1 = "Grim Reaper 1"
    InterceptorsHold = "Interceptor's Hold"
    InterceptorsHoldChests = "Interceptor's Hold Chests"
    SeadriftKeep = "Seadrift Keep"
    SeadriftKeepChests = "Seadrift Keep Chests"
    SeadriftRow = "Seadrift Row"
    SeadriftRowChests = "Seadrift Row Chests"
    CursedMedallionPopup = "Cursed Medallion Popup"
    GrimReaper2 = "Grim Reaper 2"
    DataLuxord = "Data Luxord"


class CheckLocation(str, Enum):
    RampartNavalMap = "Rampart Naval Map"
    RampartMythrilStone = "Rampart Mythril Stone"
    RampartDarkShard = "Rampart Dark Shard"
    TownDarkStone = "Town Dark Stone"
    TownApBoost = "Town AP Boost"
    TownMythrilShard = "Town Mythril Shard"
    TownMythrilGem = "Town Mythril Gem"
    CaveMouthBrightShard = "Cave Mouth Bright Shard"
    CaveMouthMythrilShard = "Cave Mouth Mythril Shard"
    IslaDeMuertaMap = "Isla de Muerta Map"
    BoatFight = "Boat Fight"
    InterceptorBarrels = "Interceptor Barrels"
    PowderStoreApBoost1 = "Powder Store AP Boost (1)"
    PowderStoreApBoost2 = "Powder Store AP Boost (2)"
    MoonlightNookMythrilShard = "Moonlight Nook Mythril Shard"
    MoonlightNookSerenityGem = "Moonlight Nook Serenity Gem"
    MoonlightNookPowerStone = "Moonlight Nook Power Stone"
    Barbossa = "Barbossa"
    FollowTheWind = "Follow the Wind"
    GrimReaper1 = "Grim Reaper 1"
    InterceptorsHoldFeatherCharm = "Interceptor's Hold Feather Charm"
    SeadriftKeepApBoost = "Seadrift Keep AP Boost"
    SeadriftKeepOrichalcum = "Seadrift Keep Orichalcum"
    SeadriftKeepMeteorStaff = "Seadrift Keep Meteor Staff"
    SeadriftRowSerenityGem = "Seadrift Row Serenity Gem"
    SeadriftRowKingRecipe = "Seadrift Row King Recipe"
    SeadriftRowMythrilCrystal = "Seadrift Row Mythril Crystal"
    SeadriftRowCursedMedallion = "Seadrift Row Cursed Medallion"
    SeadriftRowShipGraveyardMap = "Seadrift Row Ship Graveyard Map"
    GrimReaper2 = "Grim Reaper 2"
    GrimReaper2SecretAnsemReport6 = "Secret Ansem Report 6"
    DataLuxordApBoost = "Luxord (Data) AP Boost"

class PRLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_pr_keyblade(inv)
        self.logic[NodeId.Rampart][NodeId.RampartChests] = keyblade_lambda
        self.logic[NodeId.PortRoyalTown][NodeId.PortRoyalTownChests] = keyblade_lambda
        self.logic[NodeId.CaveMouth][NodeId.CaveMouthChests] = keyblade_lambda
        self.logic[NodeId.PowderStore][NodeId.PowderStoreChests] = keyblade_lambda
        self.logic[NodeId.MoonlightNook][NodeId.MoonlightNookChests] = keyblade_lambda
        self.logic[NodeId.InterceptorsHold][NodeId.InterceptorsHoldChests] = keyblade_lambda
        self.logic[NodeId.SeadriftKeep][NodeId.SeadriftKeepChests] = keyblade_lambda
        self.logic[NodeId.SeadriftRow][NodeId.SeadriftRowChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.Rampart] = ItemPlacementHelpers.pr1_check
            self.logic[NodeId.Barbossa][NodeId.GrimReaper1] = ItemPlacementHelpers.pr2_check
        else:
            self.logic[START_NODE][NodeId.Rampart] = ItemPlacementHelpers.pr1_check
            self.logic[NodeId.GrimReaper2][NodeId.PortRoyalTown] = ItemPlacementHelpers.pr2_check
            

def make_graph(graph: LocationGraphBuilder):
    pr = locationType.PR
    pr_logic = PRLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(pr_logic)

    rampart_chests = graph.add_location(NodeId.RampartChests, [
        chest(70, CheckLocation.RampartNavalMap, pr),
        chest(219, CheckLocation.RampartMythrilStone, pr),
        chest(220, CheckLocation.RampartDarkShard, pr),
    ])
    rampart = graph.add_location(NodeId.Rampart, [])
    town_chests = graph.add_location(NodeId.PortRoyalTownChests, [
        chest(71, CheckLocation.TownDarkStone, pr),
        chest(72, CheckLocation.TownApBoost, pr),
        chest(73, CheckLocation.TownMythrilShard, pr),
        chest(221, CheckLocation.TownMythrilGem, pr),
    ])
    town = graph.add_location(NodeId.PortRoyalTown, [])
    cave_mouth_chests = graph.add_location(NodeId.CaveMouthChests, [
        chest(74, CheckLocation.CaveMouthBrightShard, pr),
        chest(223, CheckLocation.CaveMouthMythrilShard, pr),
    ])
    cave_mouth = graph.add_location(NodeId.CaveMouth, [])
    isla_de_muerta_popup = graph.add_location(NodeId.IslaDeMuertaPopup, [
        popup(329, CheckLocation.IslaDeMuertaMap, pr),
    ])
    boat_fight = graph.add_location(NodeId.BoatFight, [
        item_bonus(62, CheckLocation.BoatFight, pr, vanilla=ability.AerialSpiral),
    ])
    barrels_minigame = graph.add_location(NodeId.BarrelsMinigame, [
        stat_bonus(39, CheckLocation.InterceptorBarrels, pr),
    ])
    powder_store_chests = graph.add_location(NodeId.PowderStoreChests, [
        chest(369, CheckLocation.PowderStoreApBoost1, pr),
        chest(370, CheckLocation.PowderStoreApBoost2, pr),
    ])
    powder_store = graph.add_location(NodeId.PowderStore, [])
    moonlight_nook_chests = graph.add_location(NodeId.MoonlightNookChests, [
        chest(75, CheckLocation.MoonlightNookMythrilShard, pr),
        chest(224, CheckLocation.MoonlightNookSerenityGem, pr),
        chest(371, CheckLocation.MoonlightNookPowerStone, pr),
    ])
    moonlight_nook = graph.add_location(NodeId.MoonlightNook, [])
    barbossa = graph.add_location(NodeId.Barbossa, [
        hybrid_bonus(21, CheckLocation.Barbossa, pr, vanilla=ability.AerialFinish),
        popup(263, CheckLocation.FollowTheWind, pr, vanilla=keyblade.FollowTheWind),
    ])
    grim_reaper_1 = graph.add_location(NodeId.GrimReaper1, [
        item_bonus(59, CheckLocation.GrimReaper1, pr, vanilla=ability.HorizontalSlash),
    ])
    interceptors_hold_chests = graph.add_location(NodeId.InterceptorsHoldChests, [
        chest(252, CheckLocation.InterceptorsHoldFeatherCharm, pr, vanilla=summon.FeatherCharm),
    ])
    interceptors_hold = graph.add_location(NodeId.InterceptorsHold, [])
    seadrift_keep_chests = graph.add_location(NodeId.SeadriftKeepChests, [
        chest(76, CheckLocation.SeadriftKeepApBoost, pr),
        chest(225, CheckLocation.SeadriftKeepOrichalcum, pr),
        chest(372, CheckLocation.SeadriftKeepMeteorStaff, pr),
    ])
    seadrift_keep = graph.add_location(NodeId.SeadriftKeep, [])
    seadrift_row_chests = graph.add_location(NodeId.SeadriftRowChests, [
        chest(77, CheckLocation.SeadriftRowSerenityGem, pr),
        chest(78, CheckLocation.SeadriftRowKingRecipe, pr),
        chest(373, CheckLocation.SeadriftRowMythrilCrystal, pr),
    ])
    seadrift_row = graph.add_location(NodeId.SeadriftRow, [])
    cursed_medallion_popup = graph.add_location(NodeId.CursedMedallionPopup, [
        popup(296, CheckLocation.SeadriftRowCursedMedallion, pr, vanilla=misc.CursedMedallion),
        popup(331, CheckLocation.SeadriftRowShipGraveyardMap, pr),
    ])
    grim_reaper_2 = graph.add_location(NodeId.GrimReaper2, [
        item_bonus(22, CheckLocation.GrimReaper2, pr, vanilla=magic.Magnet),
        popup(530, CheckLocation.GrimReaper2SecretAnsemReport6, pr, vanilla=report.AnsemReport6),
    ])
    data_luxord = graph.add_location(NodeId.DataLuxord, [
        popup(557, CheckLocation.DataLuxordApBoost, [pr, locationType.DataOrg]),
    ])

    graph.register_superboss(data_luxord)
    
    graph.add_edge(rampart, rampart_chests)
    graph.add_edge(town, town_chests)
    graph.add_edge(cave_mouth, cave_mouth_chests)
    graph.add_edge(powder_store, powder_store_chests)
    graph.add_edge(moonlight_nook, moonlight_nook_chests)
    graph.add_edge(interceptors_hold, interceptors_hold_chests)
    graph.add_edge(seadrift_keep, seadrift_keep_chests)
    graph.add_edge(seadrift_row, seadrift_row_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, rampart)
        graph.add_edge(rampart, town, RequirementEdge(battle=True))
        graph.add_edge(town, cave_mouth)
        graph.add_edge(cave_mouth, isla_de_muerta_popup, RequirementEdge(battle=True))
        graph.add_edge(isla_de_muerta_popup, boat_fight, RequirementEdge(battle=True))
        graph.add_edge(boat_fight, barrels_minigame)
        graph.add_edge(barrels_minigame, powder_store)
        graph.add_edge(powder_store, moonlight_nook)
        graph.add_edge(moonlight_nook, barbossa, RequirementEdge(battle=True))
        graph.add_edge(barbossa, grim_reaper_1, RequirementEdge(battle=True))
        graph.add_edge(grim_reaper_1, interceptors_hold)
        graph.add_edge(interceptors_hold, seadrift_row)
        graph.add_edge(seadrift_row, seadrift_keep)
        graph.add_edge(seadrift_keep, cursed_medallion_popup, RequirementEdge(battle=True))
        graph.add_edge(cursed_medallion_popup, grim_reaper_2, RequirementEdge(battle=True))
        graph.add_edge(grim_reaper_2, data_luxord, RequirementEdge(battle=True))
        graph.register_first_boss(barbossa)
        graph.register_last_story_boss(grim_reaper_2)
    else:
        graph.add_edge(START_NODE, rampart)
        graph.add_edge(rampart, grim_reaper_1, RequirementEdge(battle=True))
        graph.add_edge(grim_reaper_1, interceptors_hold)
        graph.add_edge(interceptors_hold, seadrift_row)
        graph.add_edge(seadrift_row, seadrift_keep)
        graph.add_edge(seadrift_keep, cursed_medallion_popup, RequirementEdge(battle=True))
        graph.add_edge(cursed_medallion_popup, cave_mouth)
        graph.add_edge(cave_mouth, grim_reaper_2, RequirementEdge(battle=True))
        graph.add_edge(grim_reaper_2, town, RequirementEdge(battle=True))
        graph.add_edge(town, isla_de_muerta_popup, RequirementEdge(battle=True))
        graph.add_edge(isla_de_muerta_popup, boat_fight, RequirementEdge(battle=True))
        graph.add_edge(boat_fight, barrels_minigame)
        graph.add_edge(barrels_minigame, powder_store)
        graph.add_edge(powder_store, moonlight_nook)
        graph.add_edge(moonlight_nook, barbossa, RequirementEdge(battle=True))
        graph.add_edge(barbossa, data_luxord, RequirementEdge(battle=True))
        graph.register_first_boss(grim_reaper_2)
        graph.register_last_story_boss(barbossa)



from enum import Enum

from List.configDict import locationType
from List.inventory import keyblade, ability, report, form, misc
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, stat_bonus, item_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    OldMansion = "Old Mansion"
    OldMansionChests = "Old Mansion Chests"
    Woods = "Woods"
    WoodsChests = "Woods Chests"
    TramCommon = "Tram Common"
    TramCommonChests = "Tram Common Chests"
    StationFightPopup = "Station Fight Popup"
    CentralStation = "TT Central Station"
    CentralStationChests = "TT Central Station Chests"
    YenSidTower = "Yensid Tower"
    YenSidTowerChests = "Yensid Tower Chests"
    YenSidTowerEntryway = "Yensid Tower Entryway"
    YenSidTowerEntrywayChests = "Yensid Tower Entryway Chests"
    SorcerersLoft = "Sorcerer's Loft"
    SorcerersLoftChests = "Sorcerer's Loft Chests"
    TowerWardrobe = "Tower Wardrobe"
    TowerWardrobeChests = "Tower Wardrobe Chests"
    ValorForm = "Valor Form"
    SeifersTrophy = "Seifer's Trophy"
    LimitForm = "Limit Form"
    UndergroundConcourse = "Underground Concourse"
    UndergroundConcourseChests = "Underground Concourse Chests"
    Tunnelway = "Tunnelway"
    TunnelwayChests = "Tunnelway Chests"
    SunsetTerrace = "TT Sunset Terrace"
    SunsetTerraceChests = "TT Sunset Terrace Chests"
    MansionBonus = "TT Mansion Bonus"
    MansionFoyer = "TT Mansion Foyer"
    MansionFoyerChests = "TT Mansion Foyer Chests"
    MansionDiningRoom = "TT Mansion Dining Room"
    MansionDiningRoomChests = "TT Mansion Dining Room Chests"
    MansionLibrary = "TT Mansion Library"
    MansionLibraryChests = "TT Mansion Library Chests"
    TwilightTownBeam = "TT Beam"
    MansionBasement = "TT Mansion Basement"
    MansionBasementChests = "TT Mansion Basement Chests"
    BetwixtAndBetween = "Betwixt and Between"
    DataAxel = "Data Axel"


class CheckLocation(str, Enum):
    OldMansionPotion = "Old Mansion Potion"
    OldMansionMythrilShard = "Old Mansion Mythril Shard"
    WoodsPotion = "The Woods Potion"
    WoodsMythrilShard = "The Woods Mythril Shard"
    WoodsHiPotion = "The Woods Hi-Potion"
    TramCommonHiPotion = "Tram Common Hi-Potion"
    TramCommonApBoost = "Tram Common AP Boost"
    TramCommonTent = "Tram Common Tent"
    TramCommonMythrilShard1 = "Tram Common Mythril Shard (1)"
    TramCommonPotion1 = "Tram Common Potion (1)"
    TramCommonMythrilShard2 = "Tram Common Mythril Shard (2)"
    TramCommonPotion2 = "Tram Common Potion (2)"
    StationPlazaSecretAnsemReport2 = "Station Plaza Secret Ansem Report 2"
    MunnyPouchMickey = "Munny Pouch (Mickey)"
    CrystalOrb = "Crystal Orb"
    CentralStationTent = "Central Station Tent"
    CentralStationHiPotion = "TT Central Station Hi-Potion"
    CentralStationMythrilShard = "Central Station Mythril Shard"
    TowerPotion = "The Tower Potion"
    TowerHiPotion = "The Tower Hi-Potion"
    TowerEther = "The Tower Ether"
    TowerEntrywayEther = "Tower Entryway Ether"
    TowerEntrywayMythrilShard = "Tower Entryway Mythril Shard"
    SorcerersLoftTowerMap = "Sorcerer's Loft Tower Map"
    TowerWardrobeMythrilStone = "Tower Wardrobe Mythril Stone"
    StarSeeker = "Star Seeker"
    ValorForm = "Valor Form"
    SeifersTrophy = "Seifer´s Trophy"
    Oathkeeper = "Oathkeeper"
    LimitForm = "Limit Form"
    UndergroundConcourseMythrilGem = "Underground Concourse Mythril Gem"
    UndergroundConcourseOrichalcum = "Underground Concourse Orichalcum"
    UndergroundConcourseApBoost = "Underground Concourse AP Boost"
    UndergroundConcourseMythrilCrystal = "Underground Concourse Mythril Crystal"
    TunnelwayOrichalcum = "Tunnelway Orichalcum"
    TunnelwayMythrilCrystal = "Tunnelway Mythril Crystal"
    SunsetTerraceOrichalcumPlus = "Sunset Terrace Orichalcum+"
    SunsetTerraceMythrilShard = "Sunset Terrace Mythril Shard"
    SunsetTerraceMythrilCrystal = "Sunset Terrace Mythril Crystal"
    SunsetTerraceApBoost = "Sunset Terrace AP Boost"
    MansionNobodies = "Mansion Nobodies"
    MansionFoyerMythrilCrystal = "Mansion Foyer Mythril Crystal"
    MansionFoyerMythrilStone = "Mansion Foyer Mythril Stone"
    MansionFoyerSerenityCrystal = "Mansion Foyer Serenity Crystal"
    MansionDiningRoomMythrilCrystal = "Mansion Dining Room Mythril Crystal"
    MansionDiningRoomMythrilStone = "Mansion Dining Room Mythril Stone"
    MansionLibraryOrichalcum = "Mansion Library Orichalcum"
    BeamSecretAnsemReport10 = "Beam Secret Ansem Report 10"
    MansionBasementCorridorUltimateRecipe = "Mansion Basement Corridor Ultimate Recipe"
    BetwixtAndBetween = "Betwixt and Between"
    BetwixtAndBetweenBondOfFlame = "Betwixt and Between Bond of Flame"
    DataAxelMagicBoost = "Axel (Data) Magic Boost"

class TTLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_tt_keyblade(inv)
        self.logic[NodeId.OldMansion][NodeId.OldMansionChests] = keyblade_lambda
        self.logic[NodeId.Woods][NodeId.WoodsChests] = keyblade_lambda
        self.logic[NodeId.TramCommon][NodeId.TramCommonChests] = keyblade_lambda
        self.logic[NodeId.CentralStation][NodeId.CentralStationChests] = keyblade_lambda
        self.logic[NodeId.UndergroundConcourse][NodeId.UndergroundConcourseChests] = keyblade_lambda
        self.logic[NodeId.Tunnelway][NodeId.TunnelwayChests] = keyblade_lambda
        self.logic[NodeId.SunsetTerrace][NodeId.SunsetTerraceChests] = keyblade_lambda
        self.logic[NodeId.MansionFoyer][NodeId.MansionFoyerChests] = keyblade_lambda
        self.logic[NodeId.MansionDiningRoom][NodeId.MansionDiningRoomChests] = keyblade_lambda
        self.logic[NodeId.MansionLibrary][NodeId.MansionLibraryChests] = keyblade_lambda
        self.logic[NodeId.MansionBasement][NodeId.MansionBasementChests] = keyblade_lambda
        # making tower checks logically need 3 ice creams to prevent softlocking
        self.logic[NodeId.YenSidTower][NodeId.YenSidTowerChests] = lambda inv: ItemPlacementHelpers.tt3_check(inv) and keyblade_lambda(inv)
        self.logic[NodeId.YenSidTowerEntryway][NodeId.YenSidTowerEntrywayChests] = lambda inv: ItemPlacementHelpers.tt3_check(inv) and keyblade_lambda(inv)
        self.logic[NodeId.SorcerersLoft][NodeId.SorcerersLoftChests] = lambda inv: ItemPlacementHelpers.tt3_check(inv) and keyblade_lambda(inv)
        self.logic[NodeId.TowerWardrobe][NodeId.TowerWardrobeChests] = lambda inv: ItemPlacementHelpers.tt3_check(inv) and keyblade_lambda(inv)
        if not reverse_rando:
            self.logic[START_NODE][NodeId.OldMansion] = ItemPlacementHelpers.tt1_check
            self.logic[NodeId.ValorForm][NodeId.SeifersTrophy] = ItemPlacementHelpers.tt2_check
            self.logic[NodeId.LimitForm][NodeId.UndergroundConcourse] = ItemPlacementHelpers.tt3_check
            self.logic[NodeId.LimitForm][NodeId.MansionBonus] = ItemPlacementHelpers.tt3_check
        else:
            self.logic[START_NODE][NodeId.CentralStation] = ItemPlacementHelpers.tt1_check
            self.logic[NodeId.BetwixtAndBetween][NodeId.SeifersTrophy] = ItemPlacementHelpers.tt2_check
            self.logic[NodeId.LimitForm][NodeId.StationFightPopup] = ItemPlacementHelpers.tt3_check

def make_graph(graph: LocationGraphBuilder):
    tt = locationType.TT
    tt_logic = TTLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(tt_logic)

    old_mansion_chests = graph.add_location(NodeId.OldMansionChests, [
        chest(447, CheckLocation.OldMansionPotion, tt),
        chest(448, CheckLocation.OldMansionMythrilShard, tt),
    ])
    old_mansion = graph.add_location(NodeId.OldMansion, [])
    woods_chests = graph.add_location(NodeId.WoodsChests, [
        chest(442, CheckLocation.WoodsPotion, tt),
        chest(443, CheckLocation.WoodsMythrilShard, tt),
        chest(444, CheckLocation.WoodsHiPotion, tt),
    ])
    woods = graph.add_location(NodeId.Woods, [])
    tram_common_chests = graph.add_location(NodeId.TramCommonChests, [
        chest(420, CheckLocation.TramCommonHiPotion, tt),
        chest(421, CheckLocation.TramCommonApBoost, tt),
        chest(422, CheckLocation.TramCommonTent, tt),
        chest(423, CheckLocation.TramCommonMythrilShard1, tt),
        chest(424, CheckLocation.TramCommonPotion1, tt),
        chest(425, CheckLocation.TramCommonMythrilShard2, tt),
        chest(484, CheckLocation.TramCommonPotion2, tt),
    ])
    tram_common = graph.add_location(NodeId.TramCommon, [])
    station_fight_popup = graph.add_location(NodeId.StationFightPopup, [
        popup(526, CheckLocation.StationPlazaSecretAnsemReport2, tt, vanilla=report.AnsemReport2),
        popup(290, CheckLocation.MunnyPouchMickey, tt, vanilla=misc.MunnyPouchMickey),
        popup(291, CheckLocation.CrystalOrb, tt),
    ])
    central_station_chests = graph.add_location(NodeId.CentralStationChests, [
        chest(431, CheckLocation.CentralStationTent, tt),
        chest(432, CheckLocation.CentralStationHiPotion, tt),
        chest(433, CheckLocation.CentralStationMythrilShard, tt),
    ])
    central_station = graph.add_location(NodeId.CentralStation, [])
    yen_sid_tower_chests = graph.add_location(NodeId.YenSidTowerChests, [
        chest(465, CheckLocation.TowerPotion, tt),
        chest(466, CheckLocation.TowerHiPotion, tt),
        chest(522, CheckLocation.TowerEther, tt),
    ])
    yen_sid_tower = graph.add_location(NodeId.YenSidTower, [])
    yen_sid_tower_entryway_chests = graph.add_location(NodeId.YenSidTowerEntrywayChests, [
        chest(467, CheckLocation.TowerEntrywayEther, tt),
        chest(468, CheckLocation.TowerEntrywayMythrilShard, tt),
    ])
    yen_sid_tower_entryway = graph.add_location(NodeId.YenSidTowerEntryway, [])
    sorcerers_loft_chests = graph.add_location(NodeId.SorcerersLoftChests, [
        chest(469, CheckLocation.SorcerersLoftTowerMap, tt),
    ])
    sorcerers_loft = graph.add_location(NodeId.SorcerersLoft, [])
    tower_wardrobe_chests = graph.add_location(NodeId.TowerWardrobeChests, [
        chest(470, CheckLocation.TowerWardrobeMythrilStone, tt),
    ])
    tower_wardrobe = graph.add_location(NodeId.TowerWardrobe, [])
    valor_form = graph.add_location(NodeId.ValorForm, [
        popup(304, CheckLocation.StarSeeker, tt, vanilla=keyblade.StarSeeker),
        popup(286, CheckLocation.ValorForm, tt, vanilla=form.ValorForm),
    ])
    seifers_trophy = graph.add_location(NodeId.SeifersTrophy, [
        popup(294, CheckLocation.SeifersTrophy, tt, vanilla=misc.SeifersTrophy),
    ])
    limit_form = graph.add_location(NodeId.LimitForm, [
        popup(265, CheckLocation.Oathkeeper, tt, vanilla=keyblade.Oathkeeper),
        popup(543, CheckLocation.LimitForm, tt, vanilla=form.LimitForm),
    ])
    underground_concourse_chests = graph.add_location(NodeId.UndergroundConcourseChests, [
        chest(479, CheckLocation.UndergroundConcourseMythrilGem, tt),
        chest(480, CheckLocation.UndergroundConcourseOrichalcum, tt),
        chest(481, CheckLocation.UndergroundConcourseApBoost, tt),
        chest(482, CheckLocation.UndergroundConcourseMythrilCrystal, tt),
    ])
    underground_concourse = graph.add_location(NodeId.UndergroundConcourse, [])
    tunnelway_chests = graph.add_location(NodeId.TunnelwayChests, [
        chest(477, CheckLocation.TunnelwayOrichalcum, tt),
        chest(478, CheckLocation.TunnelwayMythrilCrystal, tt),
    ])
    tunnelway = graph.add_location(NodeId.Tunnelway, [])
    sunset_terrace_chests = graph.add_location(NodeId.SunsetTerraceChests, [
        chest(438, CheckLocation.SunsetTerraceOrichalcumPlus, tt),
        chest(439, CheckLocation.SunsetTerraceMythrilShard, tt),
        chest(440, CheckLocation.SunsetTerraceMythrilCrystal, tt),
        chest(441, CheckLocation.SunsetTerraceApBoost, tt),
    ])
    sunset_terrace = graph.add_location(NodeId.SunsetTerrace, [])
    mansion_bonus = graph.add_location(NodeId.MansionBonus, [
        stat_bonus(56, CheckLocation.MansionNobodies, tt),
    ])
    mansion_foyer_chests = graph.add_location(NodeId.MansionFoyerChests, [
        chest(452, CheckLocation.MansionFoyerMythrilCrystal, tt),
        chest(453, CheckLocation.MansionFoyerMythrilStone, tt),
        chest(454, CheckLocation.MansionFoyerSerenityCrystal, tt),
    ])
    mansion_foyer = graph.add_location(NodeId.MansionFoyer, [])
    mansion_dining_room_chests = graph.add_location(NodeId.MansionDiningRoomChests, [
        chest(457, CheckLocation.MansionDiningRoomMythrilCrystal, tt),
        chest(458, CheckLocation.MansionDiningRoomMythrilStone, tt),
    ])
    mansion_dining_room = graph.add_location(NodeId.MansionDiningRoom, [])
    mansion_library_chests = graph.add_location(NodeId.MansionLibraryChests, [
        chest(460, CheckLocation.MansionLibraryOrichalcum, tt),
    ])
    mansion_library = graph.add_location(NodeId.MansionLibrary, [])
    beam = graph.add_location(NodeId.TwilightTownBeam, [
        popup(534, CheckLocation.BeamSecretAnsemReport10, tt, vanilla=report.AnsemReport10),
    ])
    mansion_basement_chests = graph.add_location(NodeId.MansionBasementChests, [
        chest(464, CheckLocation.MansionBasementCorridorUltimateRecipe, tt),
    ])
    mansion_basement = graph.add_location(NodeId.MansionBasement, [])
    betwixt_and_between = graph.add_location(NodeId.BetwixtAndBetween, [
        item_bonus(63, CheckLocation.BetwixtAndBetween, tt, vanilla=ability.Slapshot),
        popup(317, CheckLocation.BetwixtAndBetweenBondOfFlame, tt, vanilla=keyblade.BondOfFlame),
    ])
    data_axel = graph.add_location(NodeId.DataAxel, [
        popup(561, CheckLocation.DataAxelMagicBoost, [tt, locationType.DataOrg]),
    ])

    graph.register_superboss(data_axel)
    
    graph.add_edge(old_mansion, old_mansion_chests)
    graph.add_edge(woods, woods_chests)
    graph.add_edge(tram_common, tram_common_chests)
    graph.add_edge(central_station, central_station_chests)
    graph.add_edge(yen_sid_tower, yen_sid_tower_chests)
    graph.add_edge(yen_sid_tower_entryway, yen_sid_tower_entryway_chests)
    graph.add_edge(sorcerers_loft, sorcerers_loft_chests)
    graph.add_edge(tower_wardrobe, tower_wardrobe_chests)
    graph.add_edge(underground_concourse, underground_concourse_chests)
    graph.add_edge(tunnelway, tunnelway_chests)
    graph.add_edge(sunset_terrace, sunset_terrace_chests)
    graph.add_edge(mansion_foyer, mansion_foyer_chests)
    graph.add_edge(mansion_dining_room, mansion_dining_room_chests)
    graph.add_edge(mansion_library, mansion_library_chests)
    graph.add_edge(mansion_basement, mansion_basement_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, old_mansion)
        graph.add_edge(old_mansion, tram_common)
        graph.add_edge(tram_common, woods)
        graph.add_edge(tram_common, station_fight_popup, RequirementEdge(battle=True))
        graph.add_edge(station_fight_popup, central_station)
        graph.add_edge(central_station, yen_sid_tower)
        graph.add_edge(yen_sid_tower, yen_sid_tower_entryway)
        graph.add_edge(yen_sid_tower_entryway, sorcerers_loft, RequirementEdge(battle=True))
        graph.add_edge(sorcerers_loft, tower_wardrobe)
        graph.add_edge(tower_wardrobe, valor_form)
        graph.add_edge(valor_form, seifers_trophy, RequirementEdge(battle=True))
        graph.add_edge(seifers_trophy, limit_form)
        graph.add_edge(limit_form, underground_concourse)
        graph.add_edge(underground_concourse, tunnelway)
        graph.add_edge(tunnelway, sunset_terrace)
        graph.add_edge(limit_form, mansion_bonus, RequirementEdge(battle=True))
        graph.add_edge(mansion_bonus, mansion_foyer)
        graph.add_edge(mansion_foyer, mansion_dining_room)
        graph.add_edge(mansion_foyer, mansion_library)
        graph.add_edge(mansion_foyer, beam)
        graph.add_edge(mansion_foyer, mansion_basement)
        graph.add_edge(beam, betwixt_and_between, RequirementEdge(battle=True))
        graph.add_edge(betwixt_and_between, data_axel, RequirementEdge(battle=True))
        graph.register_first_boss(valor_form)
        graph.register_last_story_boss(betwixt_and_between)
    else:
        graph.add_edge(START_NODE, central_station)
        graph.add_edge(central_station, underground_concourse)
        graph.add_edge(central_station, tram_common)
        graph.add_edge(tram_common, woods)
        graph.add_edge(underground_concourse, tunnelway)
        graph.add_edge(tunnelway, sunset_terrace)
        graph.add_edge(woods, mansion_bonus, RequirementEdge(battle=True))
        graph.add_edge(mansion_bonus, old_mansion)
        graph.add_edge(mansion_bonus, mansion_foyer)
        graph.add_edge(mansion_foyer, mansion_dining_room)
        graph.add_edge(mansion_foyer, mansion_library)
        graph.add_edge(mansion_foyer, beam)
        graph.add_edge(mansion_foyer, mansion_basement)
        graph.add_edge(beam, betwixt_and_between, RequirementEdge(battle=True))
        graph.add_edge(betwixt_and_between, seifers_trophy,RequirementEdge(battle=True))
        graph.add_edge(seifers_trophy, limit_form)
        graph.add_edge(limit_form, station_fight_popup,RequirementEdge(battle=True))
        graph.add_edge(station_fight_popup, yen_sid_tower)
        graph.add_edge(yen_sid_tower, yen_sid_tower_entryway)
        graph.add_edge(yen_sid_tower_entryway, sorcerers_loft, RequirementEdge(battle=True))
        graph.add_edge(sorcerers_loft, tower_wardrobe)
        graph.add_edge(tower_wardrobe, valor_form)
        graph.add_edge(valor_form, data_axel, RequirementEdge(battle=True))
        graph.register_first_boss(betwixt_and_between)
        graph.register_last_story_boss(valor_form)



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Map(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.MAP)


CastleMap = Map(90, "Castle Map")
BasementMap = Map(91, "Basement Map")
CastleWallsMap = Map(92, "Castle Walls Map")
EncampmentAreaMap = Map(112, "Encampment Area Map")
VillageAreaMap = Map(113, "Village Area Map")
CornerstoneHillMap = Map(114, "Cornerstone Hill Map")
LilliputMap = Map(116, "Lilliput Map")
BuildingSiteMap = Map(117, "Building Site Map")
MickeysHouseMap = Map(118, "Mickey's House Map")
DisneyCastleMap = Map(119, "Disney Castle Map")
AgrabahMap = Map(120, "Agrabah Map")
CaveOfWondersMap = Map(121, "Cave of Wonders Map")
RuinsMap = Map(122, "Ruins Map")
UnderseaKingdomMap = Map(123, "Undersea Kingdom Map")
StarryHillMap = Map(124, "Starry Hill Map")
HundredAcreWoodMap = Map(125, "100 Acre Wood Map")
RabbitsHowseMap = Map(126, "Rabbit's House Map")
PigletsHowseMap = Map(127, "Piglet's House Map")
KangasHowseMap = Map(128, "Kanga's House Map")
SpookyCaveMap = Map(129, "Spooky Cave Map")
PalaceMap = Map(130, "Palace Map")
ColiseumMap = Map(134, "Coliseum Map")
UnderworldMap = Map(135, "Underworld Map")
CavernsMap = Map(136, "Caverns Map")
HalloweenTownMap = Map(250, "Halloween Town Map")
NavalMap = Map(251, "Naval Map")
PrideRockMap = Map(252, "Pride Rock Map")
MarketplaceMap = Map(253, "Marketplace Map")
PitCellAreaMap = Map(254, "Pit Cell Area Map")
TwilightTownMap = Map(255, "Twilight Town Map")
DarkCityMap = Map(256, "Dark City Map")
InterceptorMap = Map(505, "The Interceptor Map")
BlackPearlMap = Map(506, "The Black Pearl Map")
IslaDeMuertaMap = Map(507, "Isla de Muerta Map")
ShipGraveyardMap = Map(508, "Ship Graveyard Map")
ChristmasTownMap = Map(509, "Christmas Town Map")
CurlyHillMap = Map(510, "Curly Hill Map")
OasisMap = Map(511, "Oasis Map")
SavannahMap = Map(512, "Savannah Map")
CastlePerimeterMap = Map(513, "Castle Perimeter Map")
GreatMawMap = Map(514, "The Great Maw Map")
IoTowerMap = Map(515, "I/O Tower Map")
CentralComputerCoreMap = Map(516, "Central Computer Core Map")
SolarSailerSimulationMap = Map(517, "Solar Sailer Simulation Map")
WindowOfTimeMap = Map(518, "Window of Time Map")
SunsetHillMap = Map(531, "Sunset Hill Map")
MansionMap = Map(532, "Mansion Map")
TowerMap = Map(533, "Tower Map")
CastleThatNeverWasMap = Map(536, "Castle That Never Was Map")
DarkRemembranceMap = Map(586, "Dark Remembrance Map")
DepthsOfRemembranceMap = Map(590, "Depths of Remembrance Map")
GardenOfAssemblageMap = Map(592, "Garden of Assemblage Map")



import random
from collections import namedtuple

from Class import settingkey
from Class.seedSettings import SeedSettings
from List.configDict import (
    BattleLevelOption,
    DisableFinalOption,
    HintType,
    expCurve,
    itemBias,
    itemDifficulty,
    itemRarity,
    locationDepth,
    locationType,
    StartingVisitMode,
)

DailyModifier = namedtuple(
    "DailyModifier", ["local_modifier", "initMod", "name", "description", "categories"]
)


def levelItUpLocal(seed_settings: SeedSettings):
    seed_settings.set(settingkey.SORA_LEVELS, "ExcludeFrom99")
    seed_settings.set(settingkey.SORA_EXP_MULTIPLIER, 10.0)


def vanillaGrowth(seed_settings: SeedSettings):
    seed_settings.set(settingkey.VALOR_EXP_MULTIPLIER, 1.0)
    seed_settings.set(settingkey.WISDOM_EXP_MULTIPLIER, 1.0)
    seed_settings.set(settingkey.LIMIT_EXP_MULTIPLIER, 1.0)
    seed_settings.set(settingkey.MASTER_EXP_MULTIPLIER, 1.0)
    seed_settings.set(settingkey.FINAL_EXP_MULTIPLIER, 1.0)
    seed_settings.set(settingkey.VALOR_EXP_CURVE, expCurve.DAWN.name)
    seed_settings.set(settingkey.WISDOM_EXP_CURVE, expCurve.DAWN.name)
    seed_settings.set(settingkey.LIMIT_EXP_CURVE, expCurve.DAWN.name)
    seed_settings.set(settingkey.MASTER_EXP_CURVE, expCurve.DAWN.name)
    seed_settings.set(settingkey.FINAL_EXP_CURVE, expCurve.DAWN.name)


def addShops(seed_settings: SeedSettings):
    seed_settings.set(settingkey.SHOP_ELIXIRS, True)
    seed_settings.set(settingkey.SHOP_BOOSTS, True)
    seed_settings.set(settingkey.SHOP_RECOVERIES, True)


def shuffleBattleLevel(seed_settings: SeedSettings):
    seed_settings.set(settingkey.BATTLE_LEVEL_RANDO, BattleLevelOption.SHUFFLE.name)


def chaosBattleLevel(seed_settings: SeedSettings):
    seed_settings.set(
        settingkey.BATTLE_LEVEL_RANDO, BattleLevelOption.RANDOM_WITHIN_RANGE.name
    )
    seed_settings.set(settingkey.BATTLE_LEVEL_RANGE, 50)


def fixedBattleLevel(seed_settings: SeedSettings):
    seed_settings.set(settingkey.BATTLE_LEVEL_RANDO, BattleLevelOption.SCALE_TO_50.name)


def cupsOn(seed_settings: SeedSettings):
    seed_settings.set(settingkey.CUPS_GIVE_XP, True)
    seed_settings.set(
        settingkey.MISC_LOCATIONS_WITH_REWARDS,
        seed_settings.get(settingkey.MISC_LOCATIONS_WITH_REWARDS)
        + [locationType.OCCups.name],
    )
    seed_settings.set(
        settingkey.STARTING_INVENTORY,
        seed_settings.get(settingkey.STARTING_INVENTORY) + [537],
    )


def corOn(seed_settings: SeedSettings):
    seed_settings.set(
        settingkey.MISC_LOCATIONS_WITH_REWARDS,
        seed_settings.get(settingkey.MISC_LOCATIONS_WITH_REWARDS)
        + [locationType.CoR.name],
    )


def lockedVisitsHard(seed_settings: SeedSettings):
    seed_settings.set(settingkey.STARTING_VISIT_MODE, StartingVisitMode.RANDOM.name)
    seed_settings.set(settingkey.STARTING_VISIT_RANDOM_MAX, 6)
    seed_settings.set(settingkey.STARTING_VISIT_RANDOM_MIN, 3)
    seed_settings.set(settingkey.WEIGHTED_UNLOCKS, itemBias.SLIGHTLY_LATE)


def blockSkips(seed_settings: SeedSettings):
    seed_settings.set(settingkey.BLOCK_COR_SKIP, True)
    seed_settings.set(settingkey.BLOCK_SHAN_YU_SKIP, True)


def turnOffWorldsLocal(worlds: list):
    def _turnOffLocal(settings: SeedSettings):
        worlds_with_rewards = settings.get(settingkey.WORLDS_WITH_REWARDS)
        for world in worlds:
            if world.name in worlds_with_rewards[0]:
                worlds_with_rewards[0].remove(world.name)

    return _turnOffLocal

def mimic_item_placement_difficulty(seed_settings: SeedSettings, difficulty: itemDifficulty):
    
    map_from_difficulty_to_bias = {}
    map_from_difficulty_to_bias[itemDifficulty.SLIGHTLY_EASY] = itemBias.SLIGHTLY_EARLY
    map_from_difficulty_to_bias[itemDifficulty.SLIGHTLY_HARD] = itemBias.SLIGHTLY_LATE
    map_from_difficulty_to_bias[itemDifficulty.HARD] = itemBias.LATE
    map_from_difficulty_to_bias[itemDifficulty.VERYHARD] = itemBias.VERY_LATE
    if difficulty in map_from_difficulty_to_bias:
        map_result = map_from_difficulty_to_bias[difficulty]
    else:
        map_result = itemBias.NOBIAS

    seed_settings.set(settingkey.WEIGHTED_FORMS, map_result)
    seed_settings.set(settingkey.WEIGHTED_MAGIC, map_result)
    seed_settings.set(settingkey.WEIGHTED_PAGES, map_result)
    seed_settings.set(settingkey.WEIGHTED_SUMMONS, map_result)
    seed_settings.set(settingkey.WEIGHTED_PROOFS, map_result)
    seed_settings.set(settingkey.WEIGHTED_PROMISE_CHARM, map_result)


def enableBossEnemy(settings: SeedSettings):
    settings.set("boss", "One to One")
    settings.set("enemy", "One to One")


def modifyShutOut(daily: DailyModifier):
    X = 3
    choices = [
        locationType.Level,
        locationType.FormLevel,
        locationType.STT,
        locationType.TT,
        locationType.HB,
        locationType.BC,
        locationType.OC,
        locationType.Agrabah,
        locationType.LoD,
        locationType.HUNDREDAW,
        locationType.PL,
        locationType.DC,
        locationType.HT,
        locationType.PR,
        locationType.SP,
        locationType.TWTNW,
    ]
    random.shuffle(choices)
    shut_out_worlds = choices[:X]
    shut_out_world_names = [l.name for l in shut_out_worlds]
    daily = daily._replace(
        description=daily.description.format(", ".join(shut_out_world_names)),
        local_modifier=daily.local_modifier(shut_out_worlds),
    )
    return daily

def spoilerHints(seed_settings: SeedSettings):
    seed_settings.set(settingkey.HINT_SYSTEM, HintType.SPOILER)
    seed_settings.set(settingkey.HINTABLE_CHECKS, seed_settings.get(settingkey.HINTABLE_CHECKS) + ["report"])

def pathHints(seed_settings: SeedSettings):
    seed_settings.set(settingkey.HINT_SYSTEM, HintType.PATH)
    seed_settings.set(settingkey.HINTABLE_CHECKS, seed_settings.get(settingkey.HINTABLE_CHECKS) + ["report"])

def keybladeLocking(seed_settings: SeedSettings):
    seed_settings.set(settingkey.KEYBLADES_LOCK_CHESTS, True)
    seed_settings.set(settingkey.HINTABLE_CHECKS, seed_settings.get(settingkey.HINTABLE_CHECKS) + ["keyblade"])

dailyModifiers = [
    DailyModifier(
        name="Level Up!",
        initMod=None,
        description="Level checks up to 99 but Sora XP multiplier set to 10x",
        categories={"xp"},
        local_modifier=levelItUpLocal,
    ),
    DailyModifier(
        name="Shut Out of Worlds",
        description="The following worlds have no unique checks and are filled with junk: {}",
        initMod=modifyShutOut,
        categories={"worlds"},
        local_modifier=lambda worlds: turnOffWorldsLocal(worlds),
    ),
    DailyModifier(
        name="Path Hints",
        initMod=None,
        description="Path Hints will guide you to the proofs",
        categories={"hints"},
        local_modifier=pathHints,
    ),
    DailyModifier(
        name="Spoiler Hints",
        initMod=None,
        description="Spoiler Hints will guide you to the proofs",
        categories={"hints"},
        local_modifier=spoilerHints,
    ),
    DailyModifier(
        name="Locked Second Visits",
        initMod=None,
        description="Visit unlocks are dispersed throughout the seed, and you'll need to find them before you can enter second visits.",
        categories={"progression"},
        local_modifier=lambda settings: settings.set(settingkey.STARTING_VISIT_MODE, StartingVisitMode.FIRST.name),
    ),
    DailyModifier(
        name="Glass Cannon",
        initMod=None,
        description="Level up stats will not include Defense Ups, all stats are Strength, Magic, and Max AP",
        categories={"stats"},
        local_modifier=lambda settings: settings.set(settingkey.SORA_DEF_RATE, 0),
    ),
    DailyModifier(
        name="Moving Quick",
        initMod=None,
        description="All Growth Abilities start at level 3",
        categories={"qol"},
        local_modifier=lambda settings: settings.set(
            settingkey.STARTING_MOVEMENT, "Level_3"
        ),
    ),
    DailyModifier(
        name="Weapons In Stock",
        initMod=None,
        description="Adds all obtainable Keyblades into the Moogle shops",
        categories={"qol"},
        local_modifier=lambda settings: settings.set(settingkey.SHOP_KEYBLADES, True),
    ),
    DailyModifier(
        name="Deal More Damage",
        initMod=None,
        description="Removes the damage cap, making all stat increases matter",
        categories={"qol"},
        local_modifier=lambda settings: settings.set(
            settingkey.REMOVE_DAMAGE_CAP, True
        ),
    ),
    DailyModifier(
        name="Beatable Seed",
        initMod=None,
        description="Seed is guaranteed to give you the three proofs, but some locations may be impossible to reach",
        categories={"access"},
        local_modifier=lambda settings: settings.set(
            settingkey.ACCESSIBILITY, "beatable"
        ),
    ),
    DailyModifier(
        name="Mini Super Bosses",
        initMod=None,
        description="Adds Absent Silhouettes and Sephiroth as possible locations",
        categories={"bosses"},
        local_modifier=lambda settings: settings.set(
            settingkey.SUPERBOSSES_WITH_REWARDS,
            [locationType.AS.name, locationType.Sephi.name],
        ),
    ),
    DailyModifier(
        name="Enter the Tournament",
        initMod=None,
        description="Cups are enabled, they give experience, and you start with full access to every cup (after beating Hydra)",
        categories={"misc"},
        local_modifier=cupsOn,
    ),
    DailyModifier(
        name="What's in this Cavern?",
        initMod=None,
        description="Cavern of Remembrance is enabled",
        categories={"misc"},
        local_modifier=corOn,
    ),
    DailyModifier(
        name="Yeet the Bear",
        initMod=None,
        description="Proof of Nonexistence will be on starry hill in 100 Acre Wood",
        categories={"proof", "worlds"},
        local_modifier=lambda settings: settings.set(settingkey.YEET_THE_BEAR, True),
    ),
    DailyModifier(
        name="Proofs on Bosses",
        initMod=None,
        description="Proofs will be on the last non-superboss of a world.",
        categories={"proof"},
        local_modifier=lambda settings: settings.set(
            settingkey.PROOF_DEPTH, locationDepth.LastStoryBoss.value
        ),
    ),
    DailyModifier(
        name="Biased Checks Early",
        initMod=None,
        description="Using the Slightly Early item bias, good stuff is twice as likely to be in first half of worlds",
        categories={"placement"},
        local_modifier=lambda settings: mimic_item_placement_difficulty(settings, itemDifficulty.SLIGHTLY_EASY),
    ),
    DailyModifier(
        name="Biased Checks Late",
        initMod=None,
        description="Using the Slightly Late item bias, good stuff is twice as likely to be in second half of worlds",
        categories={"placement"},
        local_modifier=lambda settings: mimic_item_placement_difficulty(settings, itemDifficulty.SLIGHTLY_HARD),
    ),
    DailyModifier(
        name="You can have 3 of those?",
        initMod=None,
        description="Randomized Stacked Abilities is turned on, changing the pool of potential abilities you can find",
        categories={"abilities"},
        local_modifier=lambda settings: settings.set(
            settingkey.ABILITY_POOL, "randomize stackable"
        ),
    ),
    DailyModifier(
        name="Where did you find these?",
        initMod=None,
        description="Moogle shops contain useful consumable items, such as drive recoveries and elixirs",
        categories={"qol"},
        local_modifier=addShops,
    ),
    DailyModifier(
        name="What if you went to Twilight Town Last?",
        initMod=None,
        description="Shuffle the battle level of worlds.",
        categories={"btlv"},
        local_modifier=shuffleBattleLevel,
    ),
    DailyModifier(
        name="All Worlds are the Same",
        initMod=None,
        description="All max battle levels are 50, and visits are scaled.",
        categories={"btlv"},
        local_modifier=fixedBattleLevel,
    ),
]

dailyHardModifiers = [
    DailyModifier(
        name="Biased Checks Even Later",
        initMod=None,
        description="Using the Late item bias, good stuff is likely to be pushed even later.",
        categories={"placement"},
        local_modifier=lambda settings: mimic_item_placement_difficulty(settings, itemDifficulty.HARD),
    ),
    DailyModifier(
        name="Biased Checks Way Later",
        initMod=None,
        description="Using the Very Late item bias, good stuff is likely to be pushed way later.",
        categories={"placement"},
        local_modifier=lambda settings: mimic_item_placement_difficulty(settings, itemDifficulty.VERYHARD),
    ),
    DailyModifier(
        name="All Super Bosses",
        initMod=None,
        description="Adds Absent Silhouettes, Sephiroth, Lingering Will, and Data Org as possible locations",
        categories={"bosses"},
        local_modifier=lambda settings: settings.set(
            settingkey.SUPERBOSSES_WITH_REWARDS,
            [
                locationType.AS.name,
                locationType.Sephi.name,
                locationType.DataOrg.name,
                locationType.LW.name,
            ],
        ),
    ),
    DailyModifier(
        name="Locked Visits (Hard Version)",
        initMod=None,
        description="Visit unlocks are dispersed in the seed, requiring you to find them to enter worlds and make progress. Also biases them late.",
        categories={"progression"},
        local_modifier=lockedVisitsHard,
    ),
    DailyModifier(
        name="No Final Form",
        initMod=None,
        description="You are unable to use final form. You can find it for more drive levels and Genie, but you can't go into the form",
        categories={"dol"},
        local_modifier=lambda settings: settings.set(
            settingkey.DISABLE_FINAL_FORM, DisableFinalOption.NO_FINAL
        ),
    ),
    DailyModifier(
        name="Keyblades are keys?",
        initMod=None,
        description="Chests from Worlds need specific keyblades acquired to open them.",
        categories={"progression"},
        local_modifier=keybladeLocking,
    ),
    DailyModifier(
        name="No Skipping :)",
        initMod=None,
        description="You can't skip into CoR or Throne Room. Get there normally :)",
        categories={"dol"},
        local_modifier=blockSkips,
    ),
    DailyModifier(
        name="Vanilla Drive Leveling",
        initMod=None,
        description="Remember leveling drives in vanilla? You will now.",
        categories={"dol"},
        local_modifier=vanillaGrowth,
    ),
    DailyModifier(
        name="Battle levels make no sense.",
        initMod=None,
        description="Battle levels are with 50 levels of original. 1-99 possible",
        categories={"btlv"},
        local_modifier=chaosBattleLevel,
    ),
]

dailyBossEnemyModifiers = [
    DailyModifier(
        name="Final Fantasy Friends",
        initMod=None,
        description="Adds Cup Bosses to the Boss Pool",
        categories={"boss_pool"},
        local_modifier=lambda settings: settings.set("cups_bosses", True),
    ),
    DailyModifier(
        name="Enemies Change Per Room",
        initMod=None,
        description="Enemies are randomized 1-1 per room",
        categories={"enemy_pool"},
        local_modifier=lambda settings: settings.set("enemy", "One to One Per Room"),
    ),
]

dailyHardBossEnemyModifiers = [
    DailyModifier(
        name="Superbosses",
        initMod=None,
        description="Adds Superbosses to the Boss Pool",
        categories={"hard_boss_pool"},
        local_modifier=lambda settings: settings.set("data_bosses", True),
    ),
]


crit_modifier = [
    DailyModifier(
        name="Critical Mode",
        initMod=None,
        description="Enables the Randomized critical Bonuses, which you must play on critical to get.",
        categories={"hard_mode_setting"},
        local_modifier=lambda settings: settings.set(
            settingkey.CRITICAL_BONUS_REWARDS, True
        ),
    ),
]

boss_enemy_modifier = [
    DailyModifier(
        name="Boss/Enemy",
        initMod=None,
        description="Enables boss and enemy randomization",
        categories={"boss_enemy_setting"},
        local_modifier=enableBossEnemy,
    ),
]


def allDailyModifiers():
    return (
        dailyModifiers
        + dailyHardModifiers
        + dailyBossEnemyModifiers
        + dailyHardBossEnemyModifiers
    )


def getDailyModifiers(date, hard_mode=False, boss_enemy=False):
    random.seed(date.strftime("%d_%m_%Y"))
    # Weekends have more modifiers
    numMods = 3 if date.isoweekday() < 5 else 5
    chosenMods = []
    usedCategories = set()
    for _ in range(numMods):
        availableMods = []
        modifiers = dailyModifiers + ([] if not hard_mode else dailyHardModifiers)
        for m in modifiers:
            # Don't have more than one modifier from the same category
            if m.categories:
                if len(m.categories.intersection(usedCategories)) > 0:
                    continue
            # Don't have the same modifier twice
            if m.name in [m.name for m in chosenMods]:
                continue
            availableMods.append(m)
        chosen = random.choice(availableMods)
        if chosen.initMod:
            chosen = chosen.initMod(
                chosen
            )  # A little strange, but the description and modifier needs to be randomly changed
        chosenMods.append(chosen)
        for c in chosen.categories:
            usedCategories.add(c)

    if hard_mode:
        chosenMods = crit_modifier + chosenMods

    if boss_enemy:
        chosenMods = boss_enemy_modifier + chosenMods

    return chosenMods
